#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_backup_and_restore.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import shutil
import socket
import sys
import tarfile
import time
from logging.handlers import RotatingFileHandler
from os import listdir
from os.path import isfile, join

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, run_splunk_search

# import Splunk libs
import splunklib.client as client

#
# Splunk Cloud certification notes: these functions create and manage archive files for the application in the following directory $SPLUNK/etc/apps/trackme/backup
# There are no options to perform any kind of file manipulations out of this application directory
#


class TrackMeHandlerBackupAndRestore_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerBackupAndRestore_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_backup_and_restore(self, request_info, **kwargs):
        response = {
            "resource_group_name": "backup_and_restore",
            "resource_group_desc": "These endpoints provide backup and restore facilities for the Kvstore collections created and managed in TrackMe, this includes the full scope of active and enabled tenants",
        }

        return {"payload": response, "status": 200}

    # List backup archive files known from the KV, add any local file that the KV wouldn't do about
    def get_backup(self, request_info, **kwargs):

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        if describe:
            response = {
                "describe": "This endpoint lists all the backup files available on the search head, files are stored in the TrackMe backup directory, it requires a GET call with no arguments.",
                "resource_desc": "Get the list of backups known to TrackMe",
                "resource_spl_example": '| trackme mode=get url="/services/trackme/v2/backup_and_restore/backup"',
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # get local server name
            server_name = socket.gethostname()

            # check backup dir existence
            if not os.path.isdir(backuproot):
                logging.info(
                    f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                )
                return {
                    "payload": {
                        "response": f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                    },
                    "status": 200,
                }

            else:
                # store files in list
                backup_files = [
                    join(backuproot, f)
                    for f in listdir(backuproot)
                    if isfile(join(backuproot, f))
                ]

                if not backup_files:
                    logging.info(
                        f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                    )
                    return {
                        "payload": {
                            "response": f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                        },
                        "status": 200,
                    }

                else:
                    # Enter the list, verify for that for each archive file we have a corresponding record in the audit collection
                    # if there is no record, then we create a replacement record
                    for backupfile in backup_files:
                        # get the file mtime
                        try:
                            backup_file_mtime = round(os.path.getmtime(backupfile))
                        except Exception as e:
                            backup_file_mtime = None
                            logging.error(
                                f'failure to retrieve the tmime of the tar file archive="{backupfile}" with exception:"{str(e)}"'
                            )

                        # get the file size
                        try:
                            tar_filesize = os.path.getsize(backupfile)
                        except Exception as e:
                            tar_filesize = None
                            logging.error(
                                f'archive="{backupfile}", failure to retrieve the size of the tar file archive="{str(e)}"'
                            )

                        # Store a record in a backup audit collection

                        # Create a message
                        status_message = "discovered"

                        # record / key
                        record = None
                        key = None

                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "backup_archive": backupfile,
                                    "server_name": server_name,
                                }
                            ]
                        }

                        # backup audit collection
                        collection_name_backup_archives_info = (
                            "kv_trackme_backup_archives_info"
                        )
                        service_backup_archives_info = client.connect(
                            owner="nobody",
                            app="trackme",
                            port=splunkd_port,
                            token=request_info.session_key,
                            timeout=300,
                        )
                        collection_backup_archives_info = (
                            service_backup_archives_info.kvstore[
                                collection_name_backup_archives_info
                            ]
                        )

                        try:
                            kvrecord = collection_backup_archives_info.data.query(
                                query=json.dumps(query_string)
                            )[0]
                            key = kvrecord.get("_key")

                        except Exception as e:
                            key = None

                        # If a record cannot be found, this backup file is not know to TrackMe currently, add a new record
                        if key is None:
                            # attempt to extract and get the backup details

                            # Attempt extraction

                            # Set backup dir
                            backupdir = os.path.join(
                                backuproot, os.path.splitext(backupfile)[0]
                            )

                            try:
                                tf = tarfile.open(backupfile)
                                tf.extractall(backupdir)
                            except Exception as e:
                                response = {
                                    "response": f'The archive name {backupfile} could not be extracted, restore cannot be processed, exception="{str(e)}"',
                                }
                                logging.error(json.dumps(response, indent=2))
                                return {"payload": response, "status": 500}

                            # store the list of json files in a list
                            collections_json_files = [
                                f
                                for f in listdir(backupdir)
                                if isfile(join(backupdir, f))
                            ]

                            # store the list of available collections in the archive
                            collections_available = []

                            # create a dictionnary
                            collections_restore_dict = {}
                            for json_file in collections_json_files:
                                # strip the extension
                                collection_name = os.path.splitext(json_file)[0]

                                # append to the list of available collections for restore
                                collections_available.append(collection_name)

                                # get the file size
                                json_file_size = os.path.getsize(
                                    os.path.join(backupdir, json_file)
                                )

                                # get the file mtime
                                json_file_mtime = round(
                                    os.path.getmtime(os.path.join(backupdir, json_file))
                                )

                                # try getting the number of records
                                try:
                                    with open(
                                        os.path.join(backupdir, json_file), "r"
                                    ) as read_content:
                                        json_file_records = len(json.load(read_content))
                                except Exception as e:
                                    json_file_records = None

                                # add to the dict
                                collections_restore_dict[collection_name] = {
                                    "file": json_file,
                                    "size": json_file_size,
                                    "mtime": json_file_mtime,
                                    "records": json_file_records,
                                }

                            # remove backup dir
                            try:
                                shutil.rmtree(backupdir)
                            except OSError as e:
                                logging.error(
                                    f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                                )

                            try:
                                # Insert the record
                                collection_backup_archives_info.data.insert(
                                    json.dumps(
                                        {
                                            "mtime": str(backup_file_mtime),
                                            "htime": str(
                                                time.strftime(
                                                    "%c",
                                                    time.localtime(backup_file_mtime),
                                                )
                                            ),
                                            "server_name": str(server_name),
                                            "status": str(status_message),
                                            "change_type": "backup archive was missing from the info collection and added by automatic discovery",
                                            "backup_archive": str(backupfile),
                                            "size": str(tar_filesize),
                                            "archive_details": json.dumps(
                                                collections_restore_dict, indent=4
                                            ),
                                        }
                                    )
                                )

                            except Exception as e:
                                logging.error(
                                    f'failed to insert a new KVstore record with exception="{str(e)}"'
                                )
                                return {
                                    "payload": {
                                        "response": f'failed to insert a new KVstore record with exception="{str(e)}"'
                                    },
                                    "status": 500,
                                }

                    # Render

                    records = collection_backup_archives_info.data.query()
                    currently_known_archives = []
                    for record in records:
                        currently_known_archives.append(record.get("backup_archive"))

                    logging.info(
                        f'TrackMe get backup files finished successfully, archive_fields="{currently_known_archives}"'
                    )
                    return {
                        "payload": collection_backup_archives_info.data.query(),
                        "status": 200,
                    }

    # Take a backup
    def post_backup(self, request_info, **kwargs):

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        if describe:
            response = {
                "describe": "This endpoint performs a backup of all TrackMe collections in a compressed tarball, this file is stored in the TrackMe backup directory, it requires a POST call with no arguments.",
                "resource_desc": "Start TrackMe backup process",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/backup_and_restore/backup"',
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # Get service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # Set timestr
            timestr = time.strftime("trackme-backup-%Y%m%d-%H%M%S")

            # Set backup dir
            backupdir = os.path.join(backuproot, str(timestr))

            # get local server name
            server_name = socket.gethostname()

            # Create the backup dir if does not exist
            if not os.path.isdir(backuproot):
                os.mkdir(backuproot)
            if not os.path.isdir(backupdir):
                os.mkdir(backupdir)

            # Set up service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )

            # to count the number of non empty collections backed up
            counter_performed = 0

            # to count the number of empty collections
            counter_empty = 0

            # retrieve the lost of collections to be backed up

            # Spawn a new search
            # Get lagging statistics from live data
            kwargs_search = {
                "app": "trackme",
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }
            searchquery = '| rest splunk_server=local "/servicesNS/nobody/trackme/storage/collections/config" | where \'eai:acl.app\'="trackme" AND disabled=0 | table title | stats values(title) as collections | tojson | fields _raw'

            # Get the results and display them using the JSONResultsReader
            try:
                # spawn the search and get the results
                reader = run_splunk_search(
                    service,
                    searchquery,
                    kwargs_search,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        collections_json = json.loads(item.get("_raw"))
                        logging.info(
                            f'successfully retrieved the list of collections to be backed up, collections_json="{collections_json}"'
                        )

            except Exception as e:
                collections_json = None
                logging.error(
                    f'failed to retrieve the list of collections to be handled with exception="{str(e)}"'
                )

            if collections_json:
                # Loop through the collections
                for collection_name in collections_json["collections"]:
                    logging.debug(f'Handling collection="{collection_name}"')

                    try:
                        collection = service.kvstore[collection_name]

                        if str(json.dumps(collection.data.query(), indent=1)) in "[]":
                            counter_empty += 1

                        else:
                            target = os.path.join(backupdir, f"{collection_name}.json")

                            try:
                                with open(target, "w") as f:
                                    f.write(
                                        json.dumps(collection.data.query(), indent=1)
                                    )
                                counter_performed += 1

                            except Exception as e:
                                logging.error(
                                    f'Exception encountered while trying to backup collection="{collection_name}", exception="{str(e)}"'
                                )
                                return {
                                    "payload": {
                                        "response": f'Exception encountered while trying to backup collection="{collection_name}", exception="{str(e)}"',
                                    },
                                    "status": 500,
                                }

                    except Exception as e:
                        logging.error(
                            f'general exception encountered while trying to access to the KVstore, exception="{str(e)}"'
                        )
                        return {
                            "payload": f'general exception encountered while trying to access to the KVstore, exception="{str(e)}"',
                            "status": 500,
                        }

                # create a tgz
                tar_name = f"{backupdir}.tgz"

                try:
                    with tarfile.open(tar_name, mode="w:gz") as archive:
                        archive.add(backupdir, arcname="")

                except Exception as e:
                    logging.error(
                        f'Exception encountered while trying to generate the tar compressed archive="{tar_name}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while trying to generate the tar compressed archive="{tar_name}", exception="{str(e)}"',
                        },
                        "status": 500,
                    }

                # Get the archive file size
                try:
                    tar_filesize = os.path.getsize(tar_name)
                except Exception as e:
                    tar_filesize = None
                    logging.error(
                        f'tenant_id="{tar_name}", failure to retrieve the size of the tar file archive="{tar_name}" with exception:"{str(e)}"'
                    )

                # store the list of json files in a list
                collections_json_files = [
                    f for f in listdir(backupdir) if isfile(join(backupdir, f))
                ]

                # store the list of available collections in the archive
                collections_available = []

                # create a dictionnary
                collections_restore_dict = {}
                for json_file in collections_json_files:

                    if os.path.isfile(os.path.join(backupdir, json_file)):

                        # strip the extension
                        collection_name = os.path.splitext(json_file)[0]

                        # append to the list of available collections for restore
                        collections_available.append(collection_name)

                        # get the file size
                        json_file_size = os.path.getsize(
                            os.path.join(backupdir, json_file)
                        )

                        # get the file mtime
                        json_file_mtime = round(
                            os.path.getmtime(os.path.join(backupdir, json_file))
                        )

                        # try getting the number of records
                        try:
                            with open(
                                os.path.join(backupdir, json_file), "r"
                            ) as read_content:
                                json_file_records = len(json.load(read_content))
                        except Exception as e:
                            json_file_records = None

                        # add to the dict
                        collections_restore_dict[collection_name] = {
                            "file": json_file,
                            "size": json_file_size,
                            "mtime": json_file_mtime,
                            "records": json_file_records,
                        }

                # remove backup dir
                try:
                    shutil.rmtree(backupdir)
                except OSError as e:
                    logging.error(
                        f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                    )

                # Store a record in a backup audit collection

                # backup audit collection
                collection_name_backup_archives_info = "kv_trackme_backup_archives_info"
                service_backup_archives_info = client.connect(
                    owner="nobody",
                    app="trackme",
                    port=splunkd_port,
                    token=request_info.session_key,
                    timeout=300,
                )
                collection_backup_archives_info = service_backup_archives_info.kvstore[
                    collection_name_backup_archives_info
                ]

                # Get time
                current_time = int(round(time.time()))

                # Create a message
                status_message = {
                    "backup_archive": str(tar_name),
                    "collections_backed_up": str(counter_performed),
                    "collections_backed_empty": str(counter_empty),
                }

                response = {
                    "mtime": str(current_time),
                    "htime": str(time.strftime("%c", time.localtime(current_time))),
                    "server_name": str(server_name),
                    "status": json.dumps(status_message, indent=2),
                    "change_type": "backup archive created",
                    "backup_archive": str(os.path.join(backupdir, tar_name)),
                    "size": tar_filesize,
                    "archive_details": json.dumps(collections_restore_dict, indent=4),
                }

                try:
                    # Insert the record
                    collection_backup_archives_info.data.insert(json.dumps(response))

                except Exception as e:
                    logging.error(
                        f'Exception encountered while attempting to insert a record in the KVstore collection="{collection_name_backup_archives_info}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while attempting to insert a record in the KVstore collection="{collection_name_backup_archives_info}", exception="{str(e)}"'
                        },
                        "status": 500,
                    }

                # Finally render a status message
                logging.info(
                    f'TrackMe backup process is terminated, status="{json.dumps(status_message, indent=4)}"'
                )
                return {"payload": {"response": response}, "status": 200}

    # Purge older backup archives based on a retention
    def delete_backup(self, request_info, **kwargs):

        describe = False
        retention_days = 7  # default to 7 days of retention if not specified

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # get local server name
        server_name = socket.gethostname()

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            # Get the mandatory backup_archive name
            try:
                retention_days = resp_dict["retention_days"]

                # convert to an integer
                try:
                    retention_days = int(retention_days)

                except Exception as e:
                    logging.error(
                        f'Invalid retention_days="{retention_days}", expecting an integer'
                    )
                    return {
                        "payload": {
                            "response": f'Invalid retention_days="{retention_days}", expecting an integer, exception="{str(e)}"',
                        },
                        "status": 500,
                    }

            except Exception as e:
                # default to 7 days
                retention_days = 7

        else:
            # body is not required in this endpoint
            describe = False

        if describe:
            response = {
                "describe": "This endpoint performs a purge of backup archive files older than x days, it requires a DELETE call with the following arguments:",
                "resource_desc": "Purge older TrackMe backup archives",
                "resource_spl_example": "| trackme url=/services/trackme/v2/backup_and_restore/backup mode=delete body=\"{'retention_days': '7'}\"",
                "options": [
                    {
                        "retention_days": "(integer) OPTIONAL: the maximal retention for backup archive files in days, if not specified defaults to 7 days",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # For reporting purposes
            purgedlist = []

            # retention_days = 7

            time_in_secs = time.time() - (retention_days * 24 * 60 * 60)
            for root, dirs, files in os.walk(backuproot, topdown=False):
                for file_ in files:
                    full_path = os.path.join(root, file_)
                    stat = os.stat(full_path)

                    if stat.st_mtime <= time_in_secs:
                        if os.path.isdir(full_path):
                            try:
                                os.rmdir(full_path)
                                purgedlist.append(full_path)

                            except Exception as e:
                                logging.error(
                                    f'Failed to delete the backup archive="{full_path}" with exception="{str(e)}"'
                                )
                                return {
                                    "payload": {
                                        "response": f'Failed to delete the backup archive="{full_path}" with exception="{str(e)}"',
                                    },
                                    "status": 500,
                                }

                        else:
                            if os.path.exists(full_path):
                                # delete
                                os.remove(full_path)
                                # add to list
                                purgedlist.append(full_path)

                                # Purge the record from the KVstore

                                # record / key
                                record = None
                                key = None

                                # Define the KV query
                                query_string = {
                                    "backup_archive": full_path,
                                }

                                # backup audit collection
                                collection_name_backup_archives_info = (
                                    "kv_trackme_backup_archives_info"
                                )
                                service_backup_archives_info = client.connect(
                                    owner="nobody",
                                    app="trackme",
                                    port=splunkd_port,
                                    token=request_info.session_key,
                                    timeout=300,
                                )
                                collection_backup_archives_info = (
                                    service_backup_archives_info.kvstore[
                                        collection_name_backup_archives_info
                                    ]
                                )

                                try:
                                    kvrecord = (
                                        collection_backup_archives_info.data.query(
                                            query=json.dumps(query_string)
                                        )[0]
                                    )
                                    key = kvrecord.get("_key")

                                except Exception as e:
                                    key = None

                                # If a record cannot be found, this backup file is not know to TrackMe currently, add a new record
                                if key is not None:
                                    try:
                                        # Delete the record
                                        collection_backup_archives_info.data.delete(
                                            json.dumps({"_key": key})
                                        )

                                    except Exception as e:
                                        logging.error(
                                            f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"'
                                        )
                                        return {
                                            "payload": {
                                                "response": f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"',
                                            },
                                            "status": 500,
                                        }

            if not len(purgedlist) > 2:
                response = {
                    "status": f"There were no backup files older than {retention_days} days to be purged"
                }

                logging.info(json.dumps(response, indent=4))
                return {"payload": response, "status": 200}

            else:
                # For each backup archive, if a record is found in the collection info, remove the record

                # backup audit collection
                collection_name_backup_archives_info = "kv_trackme_backup_archives_info"
                service_backup_archives_info = client.connect(
                    owner="nobody",
                    app="trackme",
                    port=splunkd_port,
                    token=request_info.session_key,
                    timeout=300,
                )
                collection_backup_archives_info = service_backup_archives_info.kvstore[
                    collection_name_backup_archives_info
                ]

                for backupfile in purgedlist:
                    key = None
                    record = None

                    # Define the KV query
                    query_string = {
                        "$and": [
                            {
                                "backup_archive": backupfile,
                                "server_name": server_name,
                            }
                        ]
                    }

                    try:
                        kvrecord = collection_backup_archives_info.data.query(
                            query=json.dumps(query_string)
                        )[0]
                        key = kvrecord.get("_key")

                    except Exception as e:
                        key = None

                    # If a record is found, it shall be purged
                    if key is not None:
                        # Remove the record
                        try:
                            collection_backup_archives_info.data.delete(
                                json.dumps({"_key": key})
                            )

                        except Exception as e:
                            logging.error(
                                f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"'
                            )
                            return {
                                "payload": {
                                    "response": f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"',
                                },
                                "status": 500,
                            }

                response = {
                    "status": f'The following archives files were purged due to days retention="{retention_days}"',
                    "backup_files": purgedlist,
                }

                logging.info(json.dumps(response, indent=4))
                return {"payload": response, "status": 200}

    # Restore collections from a backup archive
    # For Splunk Cloud certification purposes, the archive must be located in the backup directory of the application

    def post_restore(self, request_info, **kwargs):

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            if not describe:
                # Get the mandatory backup_archive name
                try:
                    backup_archive = resp_dict["backup_archive"]
                except Exception as e:
                    response = {
                        "response": "The backup_archive name is required in the body of the request"
                    }
                    logging.error(json.dumps(response, indent=2))
                    return {"payload": response, "status": 500}

                # Dry run mode, preview and verify the restore operation, but do not do anything
                # default to false
                try:
                    dry_run = resp_dict["dry_run"]
                    if dry_run in ("True", "true"):
                        dry_run = True
                    else:
                        dry_run = False
                except Exception as e:
                    dry_run = True

                # Restore target, valid option are all (default) or the name of one of the KVstore collections
                # default to all if not specified
                try:
                    target = resp_dict["target"]
                    if target == "all":
                        restore_all = True
                    else:
                        restore_all = False
                        # convert as a list
                        target = target.split(",")
                except Exception as e:
                    restore_all = True

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint performs a restore of all TrackMe collections from a compressed tarball backup file"
                + "file stored in the backup directory ($SPLUNK_HOME/etc/apps/trackme/backup) of the application, it requires a POST call with thre following arguments:",
                "resource_desc": "Purge older TrackMe backup archives",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/backup_and_restore/restore\" mode=\"post\" body=\"{'backup_archive': 'trackme-backup-20210205-142635.tgz', 'dry_run': 'false', 'target': 'all'}\"",
                "options": [
                    {
                        "dry_run": "(true / false) OPTIONAL: if true, the endpoint will only verify that the archive can be found and successfully extracted, there will be no modifications at all. (default to true)",
                        "target": "(all / comma separated list of KVstore collections) OPTIONAL: restore all collections that were backed up in the submitted archive, or provide a comma separated list of collections to be restored (defaults to all)",
                        "backup_archive": "The archive file to be restoring from, the tarball compressed file must be located in the backup directory of the trackMe application.",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # Get service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # Set the submitted full path of the archive file
            backupfile = os.path.join(backuproot, backup_archive)

            # Set the full path of the directory for the extraction
            backupdir = os.path.splitext(backupfile)[0]

            # First, check the backup archive existence
            if not os.path.isfile(backupfile):
                response = {
                    "response": f"The archive name {backupfile} could not be found on the file-system, restore cannot be processed",
                }
                logging.error(json.dumps(response, indent=2))
                return {"payload": response, "status": 500}

            # Attempt extraction
            try:
                tf = tarfile.open(backupfile)
                tf.extractall(backupdir)
            except Exception as e:
                response = {
                    "response": f"The archive name {backupfile} could not be extracted, restore cannot be processed, exception='{str(e)}'",
                }
                logging.error(json.dumps(response, indent=2))
                return {"payload": response, "status": 500}

            # store the list of json files in a list
            collections_json_files = [
                f for f in listdir(backupdir) if isfile(join(backupdir, f))
            ]

            # store the list of available collections in the archive
            collections_available = []

            # create a dictionnary
            collections_restore_dict = {}
            for json_file in collections_json_files:
                # strip the extension
                collection_name = os.path.splitext(json_file)[0]

                # append to the list of available collections for restore
                collections_available.append(collection_name)

                # get the file size
                json_file_size = os.path.getsize(os.path.join(backupdir, json_file))

                # get the file mtime
                json_file_mtime = round(
                    os.path.getmtime(os.path.join(backupdir, json_file))
                )

                # try getting the number of records
                try:
                    with open(os.path.join(backupdir, json_file), "r") as read_content:
                        json_file_records = len(json.load(read_content))
                except Exception as e:
                    json_file_records = None

                # add to the dict
                collections_restore_dict[collection_name] = {
                    "file": json_file,
                    "size": json_file_size,
                    "mtime": json_file_mtime,
                    "records": json_file_records,
                }

            # if dry run
            if dry_run:
                # remove backup dir
                try:
                    shutil.rmtree(backupdir)
                except OSError as e:
                    logging.error(
                        f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                    )

                response = {
                    "response": f"Success, the archive {backupfile} could be successfully extracted, the following KVstore collections can be restored (empty collections are not backed up)",
                    "collections_json_files": collections_json_files,
                    "collections_restore_dict": collections_restore_dict,
                }

                logging.info(json.dumps(response, indent=2))
                return {"payload": response, "status": 200}

            else:
                # Set up service
                service = client.connect(
                    owner="nobody",
                    app="trackme",
                    port=splunkd_port,
                    token=request_info.session_key,
                    timeout=300,
                )

                # create a dict to store the restore results
                restore_results_dict = {}

                # create global counters
                global_records_to_be_restored = 0
                global_records_restored = 0

                # store collections with failures in a specific list
                collections_restored_warning = []

                # Things are serious now, let's restore collection per collection

                # set the list of collections to be restored
                collections_to_restored = []

                # if restore_all, add every collection that can be restored
                if restore_all:
                    # Loop through the collections
                    for collection_name in collections_restore_dict:
                        collections_to_restored.append(collection_name)

                else:
                    # Loop and check if the requested collection is available for restore
                    for collection_name in target:
                        if not collection_name in collections_available:
                            # an impossible operaton is requested, we cannot proceed
                            response = {
                                "action": "failure",
                                "response": f'the collection="{collection_name}" requested for restoration is not available in the backup archive file, restore cannot be procceded',
                                "collections_available": collections_available,
                                "collections_restore_requested": target,
                            }

                            return {"payload": response, "status": 500}

                        else:
                            collections_to_restored.append(collection_name)

                #
                # Proceed
                #

                # Loop through the restorable collections, and proceed
                for collection_name in collections_to_restored:
                    # get the file name
                    source_json_file = collections_restore_dict[collection_name]["file"]

                    # get the file size
                    source_json_file_size = collections_restore_dict[collection_name][
                        "size"
                    ]

                    # get the number of records in that file
                    source_json_file_records = collections_restore_dict[
                        collection_name
                    ]["records"]

                    # log
                    logging.info(
                        f'TrackMe restore process starting, processing restore of collection_name="{collection_name}", source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", source_json_file_records="{source_json_file_records}"'
                    )

                    # connect to collection
                    collection = service.kvstore[collection_name]

                    # try loading the json data
                    try:
                        f = open(os.path.join(backupdir, source_json_file), "r")
                        data = json.loads(f.read())

                        # add to the global counter
                        global_records_to_be_restored += len(data)

                        # process by chunk
                        counter = 0
                        chunks = [data[i : i + 500] for i in range(0, len(data), 500)]
                        for chunk in chunks:
                            # Restore from the json file
                            try:
                                collection.data.batch_save(*chunk)
                                # increment global counter restored
                                global_records_restored += len(chunk)
                                # increment counter for this collection restore operation
                                counter += len(chunk)
                            except Exception as e:
                                response = {
                                    "response": f'Restore failed while processing file="{source_json_file}" for collection="{collection_name}", exception="{str(e)}"'
                                }
                                logging.error(json.dumps(response, indent=4))
                                collections_restored_warning.append[collection_name]

                        # clean out
                        f.close()

                        # add to the results dict
                        restore_results_dict[collection_name] = {
                            "source_json_file": source_json_file,
                            "source_json_file_size": source_json_file_size,
                            "source_json_file_records": source_json_file_records,
                            "restored_records": counter,
                        }

                        # log
                        if int(source_json_file_records) != int(counter):
                            logging.error(
                                f'TrackMe restore process finished but could not verify the number of restored records to equal the number of records in the source file, processing restore of collection_name="{collection_name}", source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", source_json_file_records="{source_json_file_records}", restored_records="{counter}"'
                            )
                        else:
                            logging.info(
                                f'TrackMe restore process finished successfully, processing restore of collection_name="{collection_name}", source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", source_json_file_records="{source_json_file_records}", restored_records="{counter}"'
                            )

                    # opening the json for reading has failed
                    except Exception as e:
                        logging.error(
                            f'failed to open json file="{os.path.join(backupdir, source_json_file)}" for reading with exception="{str(e)}"'
                        )

                        # add to the results dict
                        restore_results_dict[collection_name] = {
                            "source_json_file": source_json_file,
                            "source_json_file_size": source_json_file_size,
                            "source_json_file_records": source_json_file_records,
                            "restored_records": 0,
                            "exception": f'failed to open json file="{os.path.join(backupdir, source_json_file)}" for reading with exception="{str(e)}"',
                        }

                        # add to the failing collections
                        collections_restored_warning.append[collection_name]

                # Finally remove the temporary directory
                try:
                    shutil.rmtree(backupdir)
                except OSError as e:
                    logging.error(
                        f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                    )

            # investigate the results
            if global_records_to_be_restored == global_records_restored:
                response = {
                    "response": "restore is operation is now complete and no errors were reported, please reload TrackMe",
                    "action": "success",
                    "total_records_to_be_restored": global_records_to_be_restored,
                    "total_records_restored": global_records_restored,
                    "backupfile": backupfile,
                    "restore_results_dict": restore_results_dict,
                }

            else:
                response = {
                    "response": "restore is operation is now complete but errors were reported, please review these issues and reload TrackMe",
                    "action": "warning",
                    "message": "one or more KVstore collection could not be fully restored",
                    "total_records_to_be_restored": global_records_to_be_restored,
                    "total_records_restored": global_records_restored,
                    "collections_restored_warning": collections_restored_warning,
                    "backupfile": backupfile,
                    "restore_results_dict": restore_results_dict,
                }

            logging.info(json.dumps(response, indent=4))
            return {"payload": response, "status": 200}
