#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_configuration.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from collections import OrderedDict
from logging.handlers import RotatingFileHandler


# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import test handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import (
    trackme_create_report,
    trackme_delete_report,
    trackme_create_kvtransform,
    trackme_delete_kvtransform,
    trackme_create_macro,
    trackme_delete_macro,
    trackme_create_kvcollection,
    trackme_delete_kvcollection,
)

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerConfiguration_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerConfiguration_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_configuration(self, request_info, **kwargs):
        response = {
            "resource_group_name": "configuration/admin",
            "resource_group_desc": "These endpoints provide various generic application level configuration capabilities (admin operations)",
        }

        return {"payload": response, "status": 200}

    # Create a Kvstore transforms with privileges escalation
    def post_create_kvtransform(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                transform_name = resp_dict["transform_name"]
                transform_fields = resp_dict["transform_fields"]
                collection_name = resp_dict["collection_name"]
                transform_acl = resp_dict["transform_acl"]
                owner = resp_dict["owner"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a KVstore transforms knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create KVstore transforms",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_kvstore_transforms" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"transform_name\\": \\"<transform_name>\\", \\"transform_fields\\": \\"<transform_fields>\\", \\"collection_name\\": \\"<collection_name>\\", \\"owner\\": \\"<owner>\\", \\"transform_acl\\": \\"<transform_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_kvtransform(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                transform_name,
                transform_fields,
                collection_name,
                owner,
                transform_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a Kvstore transforms with privileges escalation
    def post_delete_kvtransform(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                transform_name = resp_dict["transform_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a KVstore transforms knowledge object, it requires a POST with the following options:",
                "resource_desc": "Delete KVstore transforms",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_kvstore_transforms" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"transform_name\\": \\"<transform_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_kvtransform(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                transform_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a Kvstore collection with privileges escalation
    def post_delete_kvcollection(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                collection_name = resp_dict["collection_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a KVstore collection, it requires a POST with the following options:",
                "resource_desc": "Delete KVstore collection",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_kvstore_collection" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"collection_name\\": \\"<collection_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_kvcollection(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                collection_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the KVstore collection, collection="{collection_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a report with privileges escalation
    def post_create_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]
                report_search = resp_dict["report_search"]
                report_properties = resp_dict["report_properties"]
                report_acl = resp_dict["report_acl"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create TrackMe report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\", \\"report_search\\": \\"<report_search>\\", \\"report_properties\\": \\"<report_properties>\\", \\"report_acl\\": \\"<report_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_report(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                report_name,
                report_search,
                report_properties,
                report_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a report with privileges escalation
    def post_delete_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create delete report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_report(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                report_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Update a report with privileges escalation
    def post_update_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]
                report_search = resp_dict["report_search"]
                # Retrieving earliest and latest time from the request
                earliest_time = resp_dict.get("earliest_time")
                latest_time = resp_dict.get("latest_time")

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows updating a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Update TrackMe report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/update_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\", \\"report_search\\": \\"<report_search>\\"}"',
            }

            return {"payload": response, "status": 200}

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # update the report
        report_current = service.saved_searches[report_name]

        try:
            update_params = {"search": report_search}
            if earliest_time is not None:
                update_params["dispatch.earliest_time"] = earliest_time
            if latest_time is not None:
                update_params["dispatch.latest_time"] = latest_time

            action_update = report_current.update(**update_params)

            return {
                "payload": {
                    "action": "success",
                    "response": "The report was updated successfully",
                    "report_name": report_name,
                },
                "status": 200,
            }

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to update the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a macro with privileges escalation
    def post_create_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]
                macro_definition = resp_dict["macro_definition"]
                macro_owner = resp_dict["macro_owner"]
                macro_acl = resp_dict["macro_acl"]
        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\", \\"macro_definition\\": \\"<macro_definition>\\", \\"macro_owner\\": \\"<macro_owner>\\", \\"macro_acl\\": \\"<macro_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_macro(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                macro_name,
                macro_definition,
                macro_owner,
                macro_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Update a macro with privileges escalation
    def post_update_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]
                macro_definition = resp_dict["macro_definition"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows updating a TrackMe macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "update TrackMe macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/update_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\", \\"macro_definition\\": \\"<macro_definition>\\"}"',
            }

            return {"payload": response, "status": 200}

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # update the macro

        macro_current = service.confs["macros"][macro_name]

        try:
            action_update = macro_current.update(definition=macro_definition)
            return {
                "payload": {
                    "action": "success",
                    "response": "The macro was updated successfully",
                    "macro_name": macro_name,
                },
                "status": 200,
            }

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to update the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a macro with privileges escalation
    def post_delete_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "Delete macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_macro(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                macro_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a KVstore collection with privileges escalation
    def post_create_kvcollection(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                collection_name = resp_dict["collection_name"]
                collection_acl = resp_dict["collection_acl"]
        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a Kvstore collection, it requires a POST with the following options:",
                "resource_desc": "Create KVstore collection",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_kvcollection" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"collection_name\\": \\"<collection_name>\\", \\"collection_acl\\": \\"<collection_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_kvcollection(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                collection_name,
                collection_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the KVstore collection, collection="{collection_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}
