#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_splk_dhm.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
import time
import requests

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel

# TrackMe splk-feeds libs
from trackme_libs_splk_feeds import (
    splk_dhm_return_entity_info,
    splk_dhm_return_searches,
)

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces

# Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkDhm_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkDhm_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_dhm(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_dhm",
            "resource_group_desc": "Endpoints specific to the splk-dhm TrackMe component (Splunk Data Hosts monitoring, read only operations)",
        }

        return {"payload": response, "status": 200}

    # Get entity info
    def post_dh_entity_info(self, request_info, **kwargs):
        """
        | trackme url="/services/trackme/v2/splk_dhm/dh_entity_info" mode="post" body="{'tenant_id': 'mytenant', 'object': 'remote|account:lab|firewall.pan.amer.design.node1'}"
        """

        # init
        object_value = None
        object_id = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "the tenant_id is required",
                        },
                        "status": 500,
                    }
                # object or object_id must be provided, but not both
                try:
                    object_value = resp_dict["object"]
                except Exception as e:
                    object_value = None
                try:
                    object_id = resp_dict["object_id"]
                except Exception as e:
                    object_id = None
                if object_value and object_id:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "only object or object_id can be specified, not both",
                        },
                        "status": 500,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint returns various information related to the entity, it requires a POST call with the following information:",
                "resource_desc": "Return the entity main information, these normalized information are used to build dynamically the various queries through the user interfaces",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dhm/dh_entity_info\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'object': 'remote|account:lab|firewall.pan.amer.design.node1'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object": "The entity identifier, you can specify the object or the object_id, but not both",
                        "object_id": "The entity key identifier, you can specify the object or the object_id, but not both",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dhm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # Define the KV query
        query_string = {}

        if object_value:
            query_string["object"] = object_value
        elif object_id:
            query_string["_key"] = object_id

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            kvrecord = collection.data.query(query=json.dumps(query_string))[0]
            key = kvrecord.get("_key")
            if not object_value:
                object_value = kvrecord.get("object")

        except Exception as e:
            key = None

            if not object_value:
                return {
                    "payload": {
                        "action": "failure",
                        "response": f'the entity with identifier="{object_id}" was not found',
                    },
                    "status": 404,
                }

            else:
                return {
                    "payload": {
                        "action": "failure",
                        "response": f'the entity with identifier="{object_value}" was not found',
                    },
                    "status": 404,
                }

        # proceed
        if key:
            # init
            response = {}

            try:
                # Get entity info
                entity_info = splk_dhm_return_entity_info(kvrecord)

                # Add
                response["entity_info"] = entity_info

                # log debug
                logging.debug(
                    'function splk_dhm_return_entity_info, entity_info="{}"'.format(
                        json.dumps(entity_info, indent=2)
                    )
                )

                # Add to dict
                entity_info["index"] = kvrecord.get("data_index")
                entity_info["sourcetype"] = kvrecord.get("data_sourcetype")
                # set the search constraint
                if entity_info["breakby_key"] != "none":
                    entity_info["search_constraint"] = (
                        f'index IN ({kvrecord.get("data_index")}) sourcetype IN ({kvrecord.get("data_sourcetype")}) {entity_info["breakby_key"]}="{entity_info["breakby_value"]}"'
                    )
                else:
                    entity_info["search_constraint"] = (
                        f'index IN ({kvrecord.get("data_index")}) sourcetype IN ({kvrecord.get("data_sourcetype")}) host="{str(object_value)}"'
                    )

                # get entity searches
                entity_searches = splk_dhm_return_searches(
                    tenant_id, object_value, entity_info
                )

                # log debug
                logging.debug(
                    'function splk_dhm_return_searches, entity_searches="{}"'.format(
                        json.dumps(entity_searches, indent=2)
                    )
                )

                # add
                entity_info["splk_dhm_raw_search"] = entity_searches.get(
                    "splk_dhm_raw_search"
                )

                # single stats: manage both splunk query and trackme metric based query
                entity_info["splk_dhm_overview_splunk_single_stats"] = (
                    entity_searches.get("splk_dhm_overview_root_search")
                )
                entity_info["splk_dhm_overview_trackme_single_stats"] = (
                    remove_leading_spaces(
                        f"""\
                            | mstats avg(trackme.splk.feeds.avg_latency_5m) as avg_latency, avg(trackme.splk.feeds.perc95_latency_5m) as perc95_latency where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="splk-dhm" object="{object_value}"
                            | appendcols [ | inputlookup trackme_dhm_tenant_{tenant_id} where object="{object_value}" | eval event_delay=now()-data_last_time_seen | table event_delay ]
                        """
                    )
                )

                # add both options for trackme and splunk
                entity_info["splk_dhm_overview_splunk_pie_root_search"] = (
                    entity_searches.get("splk_dhm_overview_pie_root_search")
                )
                entity_info["splk_dhm_overview_trackme_pie_root_search"] = (
                    f'| mstats latest(trackme.splk_dhm.last_eventcount) as count where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="splk-dhm" object="{object_value}" by object, idx, st | rename idx as index, st as sourcetype'
                )

                # timechart view
                entity_info["splk_dhm_overview_splunk_timechart"] = entity_searches.get(
                    "splk_dhm_overview_timechart"
                )
                entity_info["splk_dhm_overview_trackme_timechart"] = (
                    remove_leading_spaces(
                        f"""\
                            | mstats avg(trackme.splk.feeds.lag_event_sec) as lag_event_sec, avg(trackme.splk.feeds.latest_eventcount_5m) as latest_eventcount_5m, avg(trackme.splk.feeds.avg_latency_5m) as avg_latency_5m where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="splk-dhm" object="{object_value}" by object span=5m
                            | timechart span=5m avg(avg_latency_5m) as avg_latency_5m, sum(latest_eventcount_5m) as latest_eventcount_5m, avg(lag_event_sec) as lag_event_sec
                        """
                    )
                )

                # add object and key
                entity_info["object"] = object_value
                entity_info["key"] = key

                #
                # Handle problematic backslashes in search_constraint: replace single backslashes by double backslashes
                #

                # for key pairs in entity_info, replace single backslashes by double backslashes
                for key, value in entity_info.items():
                    if isinstance(value, str):
                        entity_info[key] = value.replace("\\", "\\\\")

                # render response
                return {"payload": entity_info, "status": 200}

            except Exception as e:
                # render response
                msg = 'An exception was encountered while processing entity get info, exception="{}"'.format(
                    str(e)
                )
                logging.error(msg)
                return {
                    "payload": {
                        "action": "failure",
                        "response": msg,
                    },
                    "status": 500,
                }

    # get the component table
    def post_dhm_get_table(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "the tenant_id is required",
                        },
                        "status": 500,
                    }

                try:
                    key_id = resp_dict["key_id"]
                    if key_id == "*":
                        key_id = None
                except Exception as e:
                    key_id = None

                try:
                    object = resp_dict["object"]
                    if object == "*":
                        object = None
                except Exception as e:
                    object = None

                # only key_id or object can be specified
                if key_id and object:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "only key_id or object can be specified, not both",
                        },
                        "status": 500,
                    }

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint retrievs the component table, it requires a POST call with the following options:",
                "resource_desc": "Get the entity table",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dhm/dhm_get_table\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'key_id': '*', 'object': '*'}\"",
                "options": [
                    {
                        "tenant_id": "(required) tenant identifier",
                        "key_id": "(optional) key id, do not specify this to match all entities",
                        "object": "(optinal) entity name, do not specify this to match all entities",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # url
        url = f"{request_info.server_rest_uri}/services/trackme/v2/component/load_component_data"

        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": f"Splunk {request_info.system_authtoken}",
            "Content-Type": "application/json",
        }

        params = {
            "tenant_id": tenant_id,
            "component": "dhm",
            "page": 1,
            "size": 0,
        }

        if key_id:
            params["filter_key"] = key_id
        elif object:
            params["object_key"] = object

        data_records = []

        # Proceed
        try:
            response = requests.get(
                url,
                headers=header,
                params=params,
                verify=False,
                timeout=300,
            )

            if response.status_code not in (200, 201, 204):
                msg = f'get component has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                return {"payload": msg, "status": 500}

            else:
                response_json = response.json()
                data = response_json.get("data", [])

                # add the data to the data_records
                for record in data:
                    data_records.append(record)

                # return
                return {"payload": data_records, "status": 200}

        except Exception as e:
            msg = f'get component has failed, exception="{str(e)}"'
            logging.error(msg)
            return {"payload": msg, "status": 500}

    # get per hosts blocklists
    def post_dh_get_host_blocklists(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                action = resp_dict["action"]
                if not action in ("enable", "disable"):
                    return {
                        "payload": "Invalid option for action, valid options are: enable | disable",
                        "status": 500,
                    }
                else:
                    if action == "enable":
                        action_value = "enabled"
                    elif action == "disable":
                        action_value = "disabled"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint retrieves current per hosts blocklists definitions for a given list of entities (object), it requires a POST call with the following information:",
                "resource_desc": "Get per hosts blocklists for a given list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dhm/dh_get_host_blocklists\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'action': 'disable', 'object_list': 'key:host|linux-srv-eu1'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_value": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = f"kv_trackme_dhm_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # loop and proceed
        return_records = []

        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]
                host_idx_blocklists = kvrecord.get("host_idx_blocklists", [])
                host_st_blocklists = kvrecord.get("host_st_blocklists", [])

                # turn as lists from CSV if defined and not already
                if host_idx_blocklists:
                    if not isinstance(host_idx_blocklists, list):
                        host_idx_blocklists = host_idx_blocklists.split(",")

                if host_st_blocklists:
                    if not isinstance(host_st_blocklists, list):
                        host_st_blocklists = host_st_blocklists.split(",")

                return_records.append(
                    {
                        "keyid": key,
                        "object": kvrecord.get("object"),
                        "host_idx_blocklists": host_idx_blocklists,
                        "host_st_blocklists": host_st_blocklists,
                    }
                )
            except:
                pass

        # render
        return {"payload": return_records, "status": 200}
