#!/usr/bin/env python
# coding=utf-8

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_general_health_manager.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)

# import trackme libs
from trackme_libs import (
    trackme_reqinfo,
)

# import trackme libs
from trackme_libs import (
    run_splunk_search,
)

# import trackme libs
from trackme_libs import (
    run_splunk_search,
)

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


@Configuration(distributed=False)
class HealthTracker(GeneratingCommand):
    """
    Get all records from an ML rules collection.

    :param collection: The collection to query.
    :return: A list of records, a dictionary of records, a list of keys.
    """

    def get_ml_rules_collection(self, collection):
        collection_records = []
        collection_records_dict = {}
        count_to_process_list = []

        end = False
        skip_tracker = 0
        while not end:
            process_collection_records = collection.data.query(skip=skip_tracker)
            if process_collection_records:
                for item in process_collection_records:
                    collection_records.append(item)
                    collection_records_dict[item.get("_key")] = (
                        item  # Add the entire item to the dictionary
                    )
                    count_to_process_list.append(item.get("_key"))
                skip_tracker += 5000
            else:
                end = True

        return collection_records, collection_records_dict, count_to_process_list

    """
    Removes an orphan Machine Learning model from the collection.

    :param component: The component name.
    :param rest_url: The REST URL to use.
    :param header: The header to use.
    :param ml_model_lookup_name: The Machine Learning model lookup name.
    :return: True if the model was removed successfully, otherwise False.
    
    """

    def remove_ml_model(self, component, rest_url, header, ml_model_lookup_name):

        logging.info(
            f'component="{component}", context="mlmodels-management", attempting to delete orphan Machine Learning lookup_name="{ml_model_lookup_name}"'
        )
        try:
            response = requests.delete(
                rest_url,
                headers=header,
                verify=False,
                timeout=300,
            )
            if response.status_code not in (
                200,
                201,
                204,
            ):
                error_msg = f'context="mlmodels-management", failure to delete ML lookup_name="{ml_model_lookup_name}", url="{rest_url}", response.status_code="{response.status_code}", response.text="{response.text}"'
                raise Exception(error_msg)
            else:
                logging.info(
                    f'context="mlmodels-management", action="success", deleted lookup_name="{ml_model_lookup_name}" successfully'
                )
                return True

        except Exception as e:
            error_msg = f'context="mlmodels-management", failure to delete ML lookup_name="{ml_model_lookup_name}" with exception="{str(e)}"'
            raise Exception(error_msg)

    """
    Reasign a Machine Learning model to the Splunk system user.

    :param model_id: The model_id to reassign.
    :param rest_url: The REST URL to use.
    :param header: The header to use.

    :return: True if the model was reassigned successfully, otherwise False.

    """

    def reassign_ml_model(self, model_id, rest_url, header):

        logging.info(
            f'context="mlmodels-management", attempting to re-assign model_id="{model_id}" to splunk-system-user'
        )

        acl_properties = {
            "sharing": "user",
            "owner": "splunk-system-user",
        }

        # proceed boolean
        proceed = False

        # before re-assigning, check if the model exist by running a GET request, if the status code is different from 2**, do not proceed and log an informational message instead
        try:
            response = requests.get(
                f"{rest_url}",
                headers=header,
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.info(
                    f'context="mlmodels-management", model_id="{model_id}" does not exist, it might have been re-assigned in the meantime, skipping re-assignment'
                )
                return False
            else:
                proceed = True
        except Exception as e:
            logging.error(
                f'context="mlmodels-management", model_id="{model_id}" failed to retrieve model, exception="{str(e)}"'
            )

        if proceed:

            try:
                response = requests.post(
                    f"{rest_url}/acl",
                    headers=header,
                    data=acl_properties,
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (
                    200,
                    201,
                    204,
                ):
                    error_msg = f'context="mlmodels-management", failure to reassign model_id="{model_id}", url="{rest_url}", response.status_code="{response.status_code}", response.text="{response.text}"'
                    raise Exception(error_msg)
                else:
                    logging.info(
                        f'context="mlmodels-management", action="success", model_id="{model_id}" reassigned successfully'
                    )
                    return True

            except Exception as e:
                logging.error(
                    f'context="mlmodels-management", action="failure", model_id="{model_id}" reassigned failed, exception="{str(e)}"'
                )
                raise Exception(str(e))

    def generate(self, **kwargs):
        if self:
            # performance counter
            global_start = time.time()

            # Get request info and set logging level
            reqinfo = trackme_reqinfo(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
            )
            log.setLevel(reqinfo["logging_level"])

            # Splunk header for REST requests
            header = {
                "Authorization": f"Splunk {self._metadata.searchinfo.session_key}",
                "Content-Type": "application/json",
            }

            logging.info(
                f'context="general_execution", trackmegeneralhealthmanager is starting now.'
            )

            # global_results_dict to store results of the execution
            global_results_dict = {}

            # Register the object summary in the vtenant collection
            collection_vtenants_name = "kv_trackme_virtual_tenants"
            collection_vtenants = self.service.kvstore[collection_vtenants_name]

            # get all vtenants records, this job is not tenant specific
            vtenant_records = collection_vtenants.data.query()

            ############################################################
            # Machine Learning related global health manager tasks
            # Goals:
            # - Inspect all ML collections, identify orphans models,
            # and reassign if necessary
            ############################################################

            # Reassignment: Ensures that all ML models are owned by splunk-system-user, amd re-assign otherwise
            # run the following search to retrieve the list of existing ML models

            # counters
            ml_models_reassigned_success_count = 0
            ml_models_reassigned_failures_count = 0

            logging.info(
                f'context="mlmodels-management", subcontext="check_models:splunk-system-user_reassignment", starting verification of ML models ownership and reassignment if necessary'
            )

            # Define the query
            search = f'| rest splunk_server=local timeout=1200 "/servicesNS/nobody/trackme/data/lookup-table-files" | search eai:acl.app="trackme" AND title="__mlspl_model_*.mlmodel" | table title, id'

            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # A list to store current ml models (filename)
            ml_models_for_reassignement_current_list = []

            # A dict to store the existing models
            ml_models_for_reassignement_dict_existing = {}

            try:
                reader = run_splunk_search(
                    self.service,
                    search,
                    kwargs_oneshot,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        ml_models_for_reassignement_current_list.append(
                            item.get("title")
                        )  # this is the model filename
                        ml_models_for_reassignement_dict_existing[item.get("title")] = {
                            "id": item.get("id")
                        }

            except Exception as e:
                logging.error(
                    f'context="mlmodels-management", failed to retrieve the list of ML models, exception="{str(e)}"'
                )

            # Loop
            for model_id in ml_models_for_reassignement_current_list:

                # reassign the model
                rest_url = ml_models_for_reassignement_dict_existing[model_id].get("id")

                logging.info(
                    f'context="mlmodels-management", attempting reassignment of model_id={model_id}"'
                )

                try:
                    reassigned_model = self.reassign_ml_model(
                        model_id,
                        rest_url,
                        header,
                    )
                    if reassigned_model:
                        ml_models_reassigned_success_count += 1
                except Exception as e:
                    ml_models_reassigned_failures_count += 1
                    logging.error(
                        f'context="mlmodels-management", failed to reassign the model, model_id="{model_id}", exception="{str(e)}"'
                    )

            ############################################################
            # Identify ML models owned by splunk-system-user
            ############################################################

            # run the following search to retrieve the list of existing ML models

            logging.info(
                f'context="mlmodels-management", subcontext="check_models:splunk-system-user_owned", starting verification of ML models ownership and reassignment if necessary'
            )

            # Define the query
            search = f'| rest splunk_server=local timeout=1200 "/servicesNS/splunk-system-user/trackme/data/lookup-table-files" | search eai:acl.app="trackme" AND title="__mlspl_model_*.mlmodel" | table title, id'

            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # A list to store current ml models (filename)
            ml_models_current_list = []

            # A dict to store the existing models
            ml_models_dict_existing = {}

            try:
                reader = run_splunk_search(
                    self.service,
                    search,
                    kwargs_oneshot,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        ml_models_current_list.append(
                            item.get("title")
                        )  # this is the model filename
                        ml_models_dict_existing[item.get("title")] = {
                            "id": item.get("id")
                        }

            except Exception as e:
                logging.error(
                    f'context="mlmodels-management", failed to retrieve the list of ML models, exception="{str(e)}"'
                )

            ############################################################
            # Identify ML models configured in TrackMe
            ############################################################

            # A list to store ml_rules_outliers_collections
            ml_rules_outliers_collections = []

            # A dict to ml models definitions
            ml_models_dict = {}

            # A list to store ml models currently configured
            ml_models_list = []

            for vtenant_record in vtenant_records:
                logging.info(
                    f'context="mlmodels-management", processing vtenant_record={json.dumps(vtenant_record, indent=2)}'
                )

                # get the tenant_id
                tenant_id = vtenant_record.get("tenant_id")

                # for component in dsm, dhm, cim, flx
                for component in ["dsm", "dhm", "cim", "flx"]:

                    # get status
                    component_status = vtenant_record.get(f"tenant_{component}_enabled")

                    # append the collection
                    if component_status == 1:
                        ml_rules_outliers_collections.append(
                            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
                        )

            # for each outliers rules collection
            for ml_rules_collection_name in ml_rules_outliers_collections:

                # connect to the collection service and retrieve the records
                ml_rules_collection = self.service.kvstore[ml_rules_collection_name]

                # extract ml_rules_tenant_id from the collection name: trackme_<component>_outliers_entity_rules_tenant_<ml_rules_tenant_id>
                ml_rules_tenant_id = ml_rules_collection_name.split("_")[-1]

                # get records
                try:
                    ml_rules_records, ml_rules_records_dict, ml_rules_records_count = (
                        self.get_ml_rules_collection(ml_rules_collection)
                    )

                    for ml_rules_record in ml_rules_records:

                        # get key
                        ml_rules_record_key = ml_rules_record.get("_key")

                        # get dictionnary entities_outliers from the field entities_outliers
                        entities_outliers = json.loads(
                            ml_rules_record.get("entities_outliers")
                        )

                        # loop trough entities_outliers, the dict key is the model_id
                        for ml_model_entity in entities_outliers:

                            ml_models_dict[ml_model_entity] = {
                                "model_id": ml_model_entity,
                                "collection_name": ml_rules_collection_name,
                                "collection_key": ml_rules_record_key,
                                "tenant_id": ml_rules_tenant_id,
                            }
                            ml_models_list.append(
                                f"__mlspl_{ml_model_entity}.mlmodel"
                            )  # this is the filename

                except Exception as e:
                    logging.error(
                        f'context="mlmodels-management", failed to retrieve the records from the collection, collection_name="{ml_rules_collection_name}", exception="{str(e)}"'
                    )

                # log
                logging.info(
                    f'context="mlmodels-management", {len(ml_models_dict)} ML models were found configured in TrackMe collections, will now start inspecting Splunk existing models.'
                )

            # log the number of currently existing models
            logging.info(
                f'context="mlmodels-management", {len(ml_models_current_list)} ML models were found in the system, starting orphan models inspection'
            )

            #
            # orphan models purge / reassign
            #

            ml_models_purged_success_count = 0
            ml_models_purged_failures_count = 0

            # for each model in ml_models_current_list, if the model is not in ml_models_list, delete it
            for model_id in ml_models_current_list:
                if model_id not in ml_models_list and not model_id == "pending":
                    # remove the model
                    rest_url = ml_models_dict_existing[model_id].get("id")

                    logging.info(
                        f'context="mlmodels-management", attempting removal of model_id={model_id}"'
                    )

                    try:
                        self.remove_ml_model("trackme", rest_url, header, model_id)
                        ml_models_purged_success_count += 1
                    except Exception as e:
                        ml_models_purged_failures_count += 1
                        logging.error(
                            f'context="mlmodels-management", failed to remove the orphan model, model_id="{model_id}", exception="{str(e)}"'
                        )

            # end context="mlmodels-management"

            # log
            logging.info(
                f'context="mlmodels-management", {ml_models_purged_success_count} orphan ML models were removed, {ml_models_purged_failures_count} orphan ML models removals failed, {ml_models_reassigned_success_count} ML models were reassigned to splunk-system-user, {ml_models_reassigned_failures_count} ML models reassignments failed'
            )

            # add to results
            global_results_dict["mlmodels_management"] = {
                "ml_models_in_system_count": len(ml_models_current_list),
                "ml_models_configured_count": len(ml_models_list),
                "ml_models_purged_success_count": ml_models_purged_success_count,
                "ml_models_purged_failures_count": ml_models_purged_failures_count,
                "ml_models_reassigned_success_count": ml_models_reassigned_success_count,
                "ml_models_reassigned_failures_count": ml_models_reassigned_failures_count,
                "result": f"{ml_models_purged_success_count} orphan ML models were removed, {ml_models_purged_failures_count} orphan ML models removals failed, {ml_models_reassigned_success_count} ML models were reassigned to splunk-system-user, {ml_models_reassigned_failures_count} ML models reassignments failed",
            }

            ############################################################
            # End Machine Learning related global health manager tasks
            ############################################################

        # yield the results
        yield_record = {
            "_time": time.time(),
            "_raw": global_results_dict,
            "results": global_results_dict,
        }

        yield yield_record

        #
        # END
        #

        logging.info(
            f'context="general_execution", trackmegeneralhealthmanager has terminated, run_time={round(time.time() - global_start, 3)}'
        )


dispatch(HealthTracker, sys.argv, sys.stdin, sys.stdout, __name__)
