#!/usr/bin/env python
# coding=utf-8

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import uuid
import threading
import hashlib
from logging.handlers import RotatingFileHandler

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    f"{splunkhome}/var/log/splunk/trackme_tracker_health.log",
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# import trackme libs
from trackme_libs import (
    trackme_reqinfo,
    trackme_register_tenant_object_summary,
    trackme_delete_tenant_object_summary,
    trackme_idx_for_tenant,
    trackme_state_event,
    trackme_register_tenant_component_summary,
)

# import trackme licensing libs
from trackme_libs_licensing import trackme_check_license

# import trackme libs
from trackme_libs import (
    trackme_report_update_enablement,
    run_splunk_search,
    trackme_gen_state,
)

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces, decode_unicode

# import trackme libs logical groups
from trackme_libs_logicalgroup import (
    get_logical_groups_collection_records,
    logical_group_remove_object_from_groups,
    logical_group_delete_group_by_name,
)

# import default vtenant account settings
from collections_data import vtenant_account_default

# import trackme libs sla
from trackme_libs_sla import trackme_sla_gen_metrics

from trackme_libs_schema import trackme_schema_format_version


@Configuration(distributed=False)
class HealthTracker(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    """
    Function to return a unique uuid which is used to trace performance run_time of each subtask.
    """

    def get_uuid(self):
        return str(uuid.uuid4())

    """
    Queries and processes records from a collection based on specific criteria.

    :param collection: The collection object to query.
    :return: Tuple containing collection records and a dictionary of records.
    """

    def get_collection_records(self, component):

        # data_records
        collection_records = []
        collection_records_objects = set()
        collection_records_keys = set()
        collection_records_dict = {}

        params = {
            "tenant_id": self.tenant_id,
            "component": component,
            "page": 1,
            "size": 0,
        }

        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": f"Splunk {self._metadata.searchinfo.session_key}",
            "Content-Type": "application/json",
        }

        # Set url
        url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/component/load_component_data"

        try:
            response = requests.get(
                url,
                headers=header,
                params=params,
                verify=False,
                timeout=300,
            )

            if response.status_code not in (200, 201, 204):
                msg = f'get component has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                raise Exception(msg)

            else:
                response_json = response.json()
                data = response_json.get("data", [])

                # add the data to the data_records
                for record in data:
                    collection_records.append(record)
                    collection_records_objects.add(record.get("object"))
                    collection_records_dict[record.get("_key")] = {
                        "object": record.get("object"),
                        "object_state": record.get("object_state"),
                        "status_message": record.get("status_message", []),
                        "anomaly_reason": record.get("anomaly_reason", []),
                    }
                    collection_records_keys.add(record.get("_key"))

            return (
                collection_records,
                collection_records_objects,
                collection_records_keys,
                collection_records_dict,
            )

        except Exception as e:
            msg = f'get component has failed, exception="{str(e)}"'
            logging.error(msg)
            raise Exception(msg)

    """
    Queries and processes records from a collection based on specific criteria.

    :param collection: The collection object to query.
    :return: Tuple containing collection records and a dictionary of records.
    """

    def get_full_collection_records(self, collection):
        collection_records = []
        collection_records_dict = {}
        collection_records_keys = set()

        end = False
        skip_tracker = 0
        while not end:
            process_collection_records = collection.data.query(skip=skip_tracker)
            if process_collection_records:
                for item in process_collection_records:
                    collection_records.append(item)
                    collection_records_dict[item.get("_key")] = (
                        item  # Add the entire item to the dictionary
                    )
                    collection_records_keys.add(item.get("_key"))
                skip_tracker += 1000
            else:
                end = True

        return collection_records, collection_records_dict, collection_records_keys

    def register_component_summary_async(
        self, session_key, splunkd_uri, tenant_id, component
    ):
        try:
            summary_register_response = trackme_register_tenant_component_summary(
                session_key,
                splunkd_uri,
                tenant_id,
                component,
            )
            logging.debug(
                f'function="trackme_register_tenant_component_summary", response="{json.dumps(summary_register_response, indent=2)}"'
            )
        except Exception as e:
            logging.error(
                f'failed to register the component summary with exception="{str(e)}"'
            )

    def generate(self, **kwargs):

        # performance counter
        start = time.time()

        # set instance_id
        instance_id = self.get_uuid()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key,
            self._metadata.searchinfo.splunkd_uri,
        )
        log.setLevel(reqinfo["logging_level"])

        # get the target index
        tenant_indexes = trackme_idx_for_tenant(
            self._metadata.searchinfo.session_key,
            self._metadata.searchinfo.splunkd_uri,
            self.tenant_id,
        )

        # get trackme release
        trackme_version = None
        app_confs = self.service.confs["app"]
        for stanza in app_confs:
            if stanza.name == "id":
                for stanzakey, stanzavalue in stanza.content.items():
                    if stanzakey == "version":
                        trackme_version = stanzavalue

        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, running trackme version="{trackme_version}"'
        )

        # set the schema_version_required
        schema_version_required = trackme_schema_format_version(trackme_version)

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # Add the session_key to the reqinfo
        reqinfo["session_key"] = session_key

        # Splunk header
        header = f"Splunk {session_key}"

        # report name for logging purposes
        report_name = f"trackme_health_tracker_tenant_{self.tenant_id}"

        # Data collection
        collection_name = "kv_trackme_virtual_tenants"
        collection = self.service.kvstore[collection_name]

        # Get the tenant KVrecord
        query_string = {
            "tenant_id": self.tenant_id,
        }
        vtenant_record = collection.data.query(query=json.dumps(query_string))[0]

        #
        # check license state
        #

        task_start = time.time()
        task_instance_id = self.get_uuid()
        task_name = "check_licensing"

        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        try:
            check_license = trackme_check_license(
                reqinfo["server_rest_uri"], session_key
            )
            license_is_valid = check_license.get("license_is_valid")
            license_subscription_class = check_license.get("license_subscription_class")
            license_active_tenants = check_license.get("license_active_tenants")
            license_active_tenants_list = check_license.get(
                "license_active_tenants_list"
            )
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, function check_license called, task_instance_id={task_instance_id}, license_is_valid="{license_is_valid}", license_subscription_class="{license_subscription_class}", license_active_tenants="{license_active_tenants}", license_active_tenants_list="{license_active_tenants_list}"'
            )

        except Exception as e:
            license_is_valid = 0
            logging.error(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, function check_license exception="{str(e)}"'
            )

        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
        )

        ##################################################################################
        # Global system verifications: verify that the relevant scheduled jobs are enabled
        ##################################################################################

        # These jobs are not tenant specifics, however we use the health tracker to ensure that
        # these are effectively enabled when at least one tenant has been created and is active

        task_start = time.time()
        task_instance_id = self.get_uuid()
        task_name = "check_global_trackers_enablement"

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        savedsearch_names = [
            "trackme_ack_expiration_tracker",
            "trackme_maintenance_mode_tracker",
            "trackme_backup_scheduler",
            "trackme_general_health_manager",
        ]

        for savedsearch_name in savedsearch_names:
            # check ack expiration tracker
            update_properties_required = False

            try:
                mysavedsearch = self.service.saved_searches[savedsearch_name]
                current_disabled = int(mysavedsearch["disabled"])
                logging.debug(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, global config check, verifying savedsearch="{mysavedsearch.name}", disabled="{current_disabled}"'
                )

                if current_disabled == 1:
                    update_properties_required = True

            except Exception as e:
                logging.error(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, global configuration verification, could not retrieve the status for {savedsearch_name}'
                )

            if update_properties_required:
                try:
                    action = trackme_report_update_enablement(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        savedsearch_name,
                        "enable",
                    )
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, global config check, enabling savedsearch="{savedsearch_name}", result="{action}"'
                    )

                except Exception as e:
                    logging.error(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, global config check, an exception was encountered while trying to enable savedsearch="{savedsearch_name}", exception="{str(e)}"'
                    )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
        )

        ##################################################################################
        # Replica orchestrator
        ##################################################################################

        # This job scheduled will automatically be enabled if we detect that at least one
        # replica tracker has been created

        task_start = time.time()
        task_instance_id = self.get_uuid()
        task_name = "replica_orchestrator"

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        # Try to get the current definition
        try:
            tenant_replica_objects = vtenant_record.get("tenant_replica_objects")

            # logging debug
            logging.debug(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, tenant_replica_objects="{tenant_replica_objects}"'
            )
        except Exception as e:
            tenant_replica_objects = None

        # only run if we have a proper replica object
        if tenant_replica_objects:
            savedsearch_names = [
                "trackme_replica_executor",
            ]

            for savedsearch_name in savedsearch_names:
                # check
                update_properties_required = False

                try:
                    mysavedsearch = self.service.saved_searches[savedsearch_name]
                    current_disabled = int(mysavedsearch["disabled"])
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, replica config check, verifying savedsearch="{mysavedsearch.name}", disabled="{current_disabled}"'
                    )

                    if current_disabled == 1:
                        update_properties_required = True

                except Exception as e:
                    logging.error(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, replica configuration verification, could not retrieve the status for {savedsearch_name}'
                    )

                if update_properties_required:
                    try:
                        action = trackme_report_update_enablement(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            savedsearch_name,
                            "enable",
                        )
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, replica config check, enabling savedsearch="{savedsearch_name}", result="{action}"'
                        )

                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, replica config check, an exception was encountered while trying to enable savedsearch="{savedsearch_name}", exception="{str(e)}"'
                        )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
        )

        ###########################################################################
        # schema update and migration: detect and migrate Virtual Tenants if needed
        ###########################################################################

        task_start = time.time()
        task_instance_id = self.get_uuid()
        task_name = "schema_upgrade"

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        from trackme_libs_schema import (
            trackme_schema_get_version,
            trackme_schema_update_version,
            trackme_schema_upgrade_2009,
            trackme_schema_upgrade_2015,
            trackme_schema_upgrade_2016,
            trackme_schema_upgrade_2020,
            trackme_schema_upgrade_2026,
            trackme_schema_upgrade_2034,
            trackme_schema_upgrade_2034_least_privileges,
            trackme_schema_upgrade_2036,
            trackme_schema_upgrade_2038,
            trackme_schema_upgrade_2043,
            trackme_schema_upgrade_2044,
            trackme_schema_upgrade_2045,
            trackme_schema_upgrade_2054,
            trackme_schema_upgrade_2064,
            trackme_schema_upgrade_2067,
            trackme_schema_upgrade_2070,
            trackme_schema_upgrade_2071,
            trackme_schema_upgrade_2072,
            trackme_schema_upgrade_2075,
            trackme_schema_upgrade_2077,
            trackme_schema_upgrade_2078,
            trackme_schema_upgrade_2083,
            trackme_schema_upgrade_2084,
            trackme_schema_upgrade_2087,
            trackme_schema_upgrade_2089,
            trackme_schema_upgrade_2090,
            trackme_schema_upgrade_2091,
            trackme_schema_upgrade_2093,
            trackme_schema_upgrade_2094,
            trackme_schema_upgrade_2095,
            trackme_schema_upgrade_2096,
            trackme_schema_upgrade_2097,
            trackme_schema_upgrade_2098,
            trackme_schema_upgrade_2099,
            trackme_schema_upgrade_2100,
            trackme_schema_upgrade_2101,
            trackme_schema_upgrade_2102,
        )

        # Define a mapping between schema versions and their upgrade functions
        schema_upgrades = [
            (2009, trackme_schema_upgrade_2009),
            (2015, trackme_schema_upgrade_2015),
            (2016, trackme_schema_upgrade_2016),
            (2020, trackme_schema_upgrade_2020),
            (2026, trackme_schema_upgrade_2026),
            (2034, trackme_schema_upgrade_2034),
            (2034, trackme_schema_upgrade_2034_least_privileges),
            (2036, trackme_schema_upgrade_2036),
            (2038, trackme_schema_upgrade_2038),
            (2043, trackme_schema_upgrade_2043),
            (2043, trackme_schema_upgrade_2044),
            (2045, trackme_schema_upgrade_2045),
            (2054, trackme_schema_upgrade_2054),
            (2064, trackme_schema_upgrade_2064),
            (2067, trackme_schema_upgrade_2067),
            (2070, trackme_schema_upgrade_2070),
            (2071, trackme_schema_upgrade_2071),
            (2072, trackme_schema_upgrade_2072),
            (2075, trackme_schema_upgrade_2075),
            (2077, trackme_schema_upgrade_2077),
            (2078, trackme_schema_upgrade_2078),
            (2083, trackme_schema_upgrade_2083),
            (2084, trackme_schema_upgrade_2084),
            (2087, trackme_schema_upgrade_2087),
            (2089, trackme_schema_upgrade_2089),
            (2090, trackme_schema_upgrade_2090),
            (2091, trackme_schema_upgrade_2091),
            (2093, trackme_schema_upgrade_2093),
            (2094, trackme_schema_upgrade_2094),
            (2095, trackme_schema_upgrade_2095),
            (2096, trackme_schema_upgrade_2096),
            (2097, trackme_schema_upgrade_2097),
            (2098, trackme_schema_upgrade_2098),
            (2099, trackme_schema_upgrade_2099),
            (2100, trackme_schema_upgrade_2100),
            (2101, trackme_schema_upgrade_2101),
            (2102, trackme_schema_upgrade_2102),
        ]

        # Get the current schema version
        try:
            schema_version = trackme_schema_get_version(
                reqinfo,
                self.tenant_id,
                schema_version_required,
                task_name,
                task_instance_id,
            )

        except Exception as e:
            logging.error(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, tenant_id="{self.tenant_id}", failed to call function trackme_schema_get_version, exception="{str(e)}"'
            )

        # Check and act accordingly
        trackme_backup_attempted = False

        if not schema_version or int(schema_version) != int(schema_version_required):
            for version, upgrade_func in schema_upgrades:
                if not schema_version or int(schema_version) < version:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, detected migration required for schema version {version}, schema_version="{schema_version}", schema_version_required="{schema_version_required}", processing now.'
                    )

                    # before running the first function, execute TrackMe's builtin backup job
                    if not trackme_backup_attempted:
                        try:
                            response = requests.post(
                                f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/backup_and_restore/backup",
                                headers={
                                    "Authorization": header,
                                    "Content-Type": "application/json",
                                },
                                verify=False,
                                timeout=300,
                            )
                            if response.status_code not in (200, 201, 204):
                                logging.error(
                                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, backup post call has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                                )
                            else:
                                logging.info(
                                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, backup post call executed successfully'
                                )
                                vtenant_account = response.json()
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, backup post call has failed, exception="{str(e)}"'
                            )
                        trackme_backup_attempted = True

                    # proceed
                    try:
                        schema_version_update = upgrade_func(
                            reqinfo,
                            self.tenant_id,
                            int(schema_version),
                            int(schema_version_required),
                            task_name,
                            task_instance_id,
                        )
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, schema version {version} migrated successfully.'
                        )
                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, failed to call function {upgrade_func.__name__}, exception="{str(e)}"'
                        )

            #
            # finally migrate the schema version
            #

            try:
                schema_version_update = trackme_schema_update_version(
                    reqinfo,
                    self.tenant_id,
                    schema_version_required,
                    task_name,
                    task_instance_id,
                )
            except Exception as e:
                logging.error(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, failed to call function trackme_schema_update_version, exception="{str(e)}"'
                )

        else:
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, schema is up to date, no action required, schema_version="{schema_version}"'
            )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        #
        #

        ###########################################################################
        # verify that the tenant account exists as expected
        ###########################################################################

        task_start = time.time()
        task_instance_id = self.get_uuid()
        task_name = "check_vtenant_accounts"

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, verifying the vtenant account'
        )

        # target
        url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/vtenants_accounts"

        # init
        vtenant_account = None

        # proceed
        try:
            response = requests.post(
                url,
                headers={
                    "Authorization": header,
                    "Content-Type": "application/json",
                },
                data=json.dumps({"tenant_id": self.tenant_id}),
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.error(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, verify vtenant account has failed, was this account deleted by mistake? response.status_code="{response.status_code}", response.text="{response.text}"'
                )
            else:
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, account was verified successfully'
                )
                vtenant_account = response.json()
        except Exception as e:
            logging.error(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, verify vtenant account has failed, exception="{str(e)}"'
            )

        # act if needed
        if not vtenant_account:
            url = f"{self._metadata.searchinfo.splunkd_uri}/servicesNS/nobody/trackme/trackme_vtenants"

            # load default vtenant config
            data = dict(vtenant_account_default)

            # add the name value
            data["name"] = self.tenant_id

            # Retrieve and set the tenant idx, if any failure, logs and use the global index
            try:
                response = requests.post(
                    url,
                    headers={
                        "Authorization": header,
                        "Content-Type": "application/json",
                    },
                    data=data,
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, create vtenant account has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, create vtenant account was operated successfully, response.status_code="{response.status_code}"'
                    )
            except Exception as e:
                logging.error(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, create vtenant account has failed, exception="{str(e)}"'
                )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        #
        #

        #
        # all components - inspect_collection
        #

        # context: this activty verifies that the collection record object statuses are consistent according to the Decision Maker
        # It works by loading the component dta, then looping trough objects to verify and update their collection status if needed

        for component in ("dsm", "dhm", "mhm", "wlk", "flx"):

            if vtenant_record.get(f"tenant_{component}_enabled") == True:
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting collection records object statuses now.'
                )

                # set collection target
                inspect_collection_name = (
                    f"kv_trackme_{component}_tenant_{self.tenant_id}"
                )
                inspect_collection = self.service.kvstore[inspect_collection_name]

                #
                # subtask: permanently_deleted_records_inspection
                #

                task_instance_id = self.get_uuid()
                task_start = time.time()
                task_name = "inspect_collection:permanently_deleted_records_inspection"

                # start task
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
                )

                #
                # Check permanently deleted records:
                # A permanently deleted record should not exist in the main KVstore collection, if it does, it should be purged
                #

                # Lists to store permanently deleted records found in anomaly
                collection_permanently_deleted_records_anomaly = []

                # search
                search = remove_leading_spaces(
                    f"""\
                    | inputlookup trackme_{component}_tenant_{self.tenant_id} | eval keyid=_key
                    | lookup trackme_common_permanently_deleted_objects_tenant_{self.tenant_id} object, object_category OUTPUT _key as permanently_deleted_keys
                    | where isnotnull(permanently_deleted_keys)
                    | table keyid, *
                """
                )

                # kwargs
                kwargs_search = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "preview": "false",
                    "output_mode": "json",
                    "count": 0,
                }

                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting the main data collection for permanently deleted records now.'
                )

                try:
                    reader = run_splunk_search(
                        self.service,
                        search,
                        kwargs_search,
                        24,
                        5,
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            collection_permanently_deleted_records_anomaly.append(item)

                except Exception as e:
                    msg = f'permanently deleted records inspection search failed with exception="{str(e)}"'
                    logging.error(msg)
                    raise Exception(msg)

                if len(collection_permanently_deleted_records_anomaly) > 0:
                    logging.warning(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, permanently deleted records found, no_records="{len(collection_permanently_deleted_records_anomaly)}"'
                    )

                    for record in collection_permanently_deleted_records_anomaly:
                        try:
                            inspect_collection.data.delete(
                                json.dumps({"_key": record.get("keyid")})
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, entities in the main collection which are also in ther permanently deleted records were purged successfully, keyid="{record.get("keyid")}", record="{json.dumps(record, indent=1)}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, failed to delete permanently deleted records in anmaly, keyid="{record.get("keyid")}", , record="{json.dumps(record, indent=1)}", exception="{str(e)}"'
                            )

                else:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, no permanenlty deleted records in anomaly found'
                    )

                # end subtask
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
                )

                #
                # subtask: corrupted_records_inspection
                #

                task_start = time.time()
                task_instance_id = self.get_uuid()
                task_name = "inspect_collection:corrupted_records_inspection"

                # start task
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
                )

                #
                # Check for unexpected corrupted record, a foreign record which has been stored in the KVstore by mistake
                # would not have an object value, and would be purged if any.
                #

                # Lists to store corrupted records
                collection_corrupted_records = []

                # search
                search = remove_leading_spaces(
                    f"""\
                    | inputlookup trackme_{component}_tenant_{self.tenant_id} | eval keyid=_key
                    | where isnull(object) OR object=""
                    | table keyid, *
                """
                )

                # kwargs
                kwargs_search = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "preview": "false",
                    "output_mode": "json",
                    "count": 0,
                }

                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting the main data collection for corrupted records now.'
                )

                try:
                    reader = run_splunk_search(
                        self.service,
                        search,
                        kwargs_search,
                        24,
                        5,
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            collection_corrupted_records.append(item)

                except Exception as e:
                    msg = f'corrupted record inspection search failed with exception="{str(e)}"'
                    logging.error(msg)
                    raise Exception(msg)

                if len(collection_corrupted_records) > 0:
                    logging.warning(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, corrupted records found, no_records="{len(collection_corrupted_records)}"'
                    )

                    for corrupted_record in collection_corrupted_records:
                        try:
                            inspect_collection.data.delete(
                                json.dumps({"_key": corrupted_record.get("keyid")})
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, corrupted record deleted successfully, keyid="{corrupted_record.get("keyid")}", record="{json.dumps(corrupted_record, indent=1)}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, failed to delete corrupted record, keyid="{corrupted_record.get("keyid")}", , record="{json.dumps(corrupted_record, indent=1)}", exception="{str(e)}"'
                            )

                else:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, no corrupted records found'
                    )

                # end subtask
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
                )

                #
                # subtask: missing_tenant_id_records_inspection
                #

                task_start = time.time()
                task_instance_id = self.get_uuid()
                task_name = "inspect_collection:missing_tenant_id_records_inspection"

                # start task
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
                )

                #
                # Check for records which would miss the tenant_id field, and add it if needed
                #

                # collection_misggin_tenant_id_records records
                collection_missing_tenant_id_records = []

                # search
                search = remove_leading_spaces(
                    f"""\
                    | inputlookup trackme_{component}_tenant_{self.tenant_id} | eval keyid=_key
                    | where isnull(tenant_id) OR tenant_id=""
                    | table keyid, *
                """
                )

                # kwargs
                kwargs_search = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "preview": "false",
                    "output_mode": "json",
                    "count": 0,
                }

                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting the main data collection for corrupted records now.'
                )

                try:
                    reader = run_splunk_search(
                        self.service,
                        search,
                        kwargs_search,
                        24,
                        5,
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            collection_missing_tenant_id_records.append(item)

                except Exception as e:
                    msg = f'missing tenant_id record inspection search failed with exception="{str(e)}"'
                    logging.error(msg)
                    raise Exception(msg)

                if len(collection_missing_tenant_id_records) > 0:
                    logging.warning(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, records found, no_records="{len(collection_missing_tenant_id_records)}"'
                    )

                    for missing_record in collection_missing_tenant_id_records:
                        try:
                            missing_record["tenant_id"] = self.tenant_id
                            inspect_collection.data.update(
                                missing_record.get("_key"),
                                json.dumps(missing_record),
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, record updated successfully, keyid="{missing_record.get("keyid")}", record="{json.dumps(missing_record, indent=1)}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, failed to update record, keyid="{missing_record.get("keyid")}", , record="{json.dumps(missing_record, indent=1)}", exception="{str(e)}"'
                            )

                else:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, no records found'
                    )

                # end subtask
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
                )

                #
                # subtask: entities_auto_disablement
                #

                task_start = time.time()
                task_instance_id = self.get_uuid()
                task_name = "inspect_collection:entities_auto_disablement"

                # start task
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
                )

                #
                # Check for feeds entities to be disabled according to the system wide setting: splk_general_feeds_auto_disablement_period
                # This setting allows to disable feeds entities if they have not been updated for a certain period of time
                #

                # system wide setting
                splk_general_feeds_auto_disablement_period = reqinfo["trackme_conf"][
                    "splk_general"
                ]["splk_general_feeds_auto_disablement_period"]

                # tenant setting (override system wide setting, if set)
                splk_feeds_auto_disablement_period = vtenant_account.get(
                    "splk_feeds_auto_disablement_period"
                )

                # handle
                auto_disablement_period = (
                    splk_feeds_auto_disablement_period
                    if splk_feeds_auto_disablement_period
                    else splk_general_feeds_auto_disablement_period
                )

                if auto_disablement_period != "0d" and component in (
                    "dsm",
                    "dhm",
                    "mhm",
                ):

                    # Lists to store entities to be disabled
                    entities_to_be_disabled = []

                    # search
                    search = remove_leading_spaces(
                        f"""\
                        | inputlookup trackme_{component}_tenant_{self.tenant_id} | eval keyid=_key
                        | eval last_time_seen=coalesce(data_last_time_seen, metric_last_time_seen)
                        | where last_time_seen<=`trackme_auto_disablement_period`
                        | table keyid, object, last_time_seen
                        | eval last_time_seen_human=strftime(last_time_seen, "%c")
                    """
                    )

                    # kwargs
                    kwargs_search = {
                        "earliest_time": "-5m",
                        "latest_time": "now",
                        "preview": "false",
                        "output_mode": "json",
                        "count": 0,
                    }

                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting the main data collection for entities to be disabled according to auto-disablement setting. (auto_disablement_period="{auto_disablement_period}")'
                    )

                    try:
                        reader = run_splunk_search(
                            self.service,
                            search,
                            kwargs_search,
                            24,
                            5,
                        )

                        for item in reader:
                            if isinstance(item, dict):
                                entities_to_be_disabled.append(item.get("keyid"))

                    except Exception as e:
                        msg = f'auto-disablement record inspection search failed with exception="{str(e)}"'
                        logging.error(msg)
                        raise Exception(msg)

                    if len(entities_to_be_disabled) > 0:
                        logging.warning(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, entities to be disabled were found, list="{entities_to_be_disabled}"'
                        )

                        # turn entities_to_be_disabled list into CSV
                        entities_to_be_disabled_csv = ",".join(entities_to_be_disabled)

                        # call mass disablement endpoint
                        if component == "dsm":
                            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dsm/write/ds_monitoring"
                        elif component == "dhm":
                            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dhm/write/dh_monitoring"
                        elif component == "mhm":
                            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_mhm/write/mh_monitoring"

                        try:
                            response = requests.post(
                                target_url,
                                headers={
                                    "Authorization": header,
                                    "Content-Type": "application/json",
                                },
                                data=json.dumps(
                                    {
                                        "tenant_id": self.tenant_id,
                                        "keys_list": entities_to_be_disabled_csv,
                                        "action": "disable",
                                        "update_comment": f"auto-disabled by the system, last seen data is beyond the system wide auto-disablement period of {splk_general_feeds_auto_disablement_period}",
                                    }
                                ),
                                verify=False,
                                timeout=300,
                            )

                            if response.status_code not in (200, 201, 204):
                                msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, query has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                                logging.error(msg)
                            else:
                                try:
                                    success_count = response.json().get("success_count")
                                except Exception as e:
                                    success_count = 0
                                msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, request was successful, success_count="{success_count}"'
                                logging.info(msg)

                        except Exception as e:
                            msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, ctask="{task_name}", task_instance_id={task_instance_id}, request failed with exception="{str(e)}"'
                            logging.info(msg)

                    else:
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, no entities to be disabled were found'
                        )

                # end subtask
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
                )

                #
                # subtask: handle_sync_entities
                #

                task_start = time.time()
                task_instance_id = self.get_uuid()
                task_name = "inspect_collection:handle_sync_entities"

                # start task
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
                )

                #
                # Inspecting statuses
                #

                #
                # START raw collections records: Get raw collection records using a Splunk search
                #

                # search
                search = remove_leading_spaces(
                    f"""\
                    | trackmegetcoll tenant_id="{self.tenant_id}" component="{component}" | fields - _raw | table *
                    | lookup trackme_{component}_tenant_{self.tenant_id} _key as keyid OUTPUT object_state as kvcoll_object_state, anomaly_reason as kvcoll_anomaly_reason, latest_flip_time as kvcoll_latest_flip_time
                    | where object_state!=kvcoll_object_state
                """
                )

                # kwargs
                kwargs_search = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "preview": "false",
                    "output_mode": "json",
                    "count": 0,
                }

                delta_records = []
                delta_records_keys = set()
                delta_records_objects = set()
                delta_records_dict = {}

                try:
                    reader = run_splunk_search(
                        self.service,
                        search,
                        kwargs_search,
                        24,
                        5,
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            delta_records.append(item)
                            delta_records_keys.add(item.get("keyid"))
                            delta_records_objects.add(item.get("object"))
                            delta_records_dict[item.get("keyid")] = item

                except Exception as e:
                    msg = f'main search failed with exception="{str(e)}"'
                    logging.error(msg)
                    raise Exception(msg)

                #
                # END raw collections records: Get raw collection records using a Splunk search
                #

                #
                # Handle delta records
                #

                inspectcollection_compare_records_start_time = time.time()

                for item in delta_records:

                    item_key = item.get("keyid")
                    item_object = decode_unicode(item.get("object"))
                    item_alias = item.get("alias")
                    item_object_state = item.get("object_state")
                    item_object_category = item.get("object_category")
                    item_anomaly_reason = item.get("anomaly_reason")
                    item_monitored_state = item.get("monitored_state")
                    item_priority = item.get("priority")

                    # our delta state
                    collection_object_state = item.get("kvcoll_object_state")

                    # previous_anomaly_reason
                    collection_anomaly_reason = item.get(
                        "kvcoll_anomaly_reason", "unknown"
                    )

                    # previous flip time
                    try:
                        collection_latest_flip_time = float(
                            item.get("kvcoll_latest_flip_time", 0)
                        )
                    except Exception as e:
                        collection_latest_flip_time = 0

                    # disruption time
                    disruption_time = 0

                    # compare the object state with item_object_state using decisionmaker_collection_records_dict using the key
                    # if the object_state value is different, log the issue

                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, collection record object state is not consistent, object="{item_object}", object_id="{item_key}", in_collection_object_state="{collection_object_state}", in_result_object_state="{item_object_state}", in_collection_anomaly_reason="{collection_anomaly_reason}"'
                    )

                    # get the current kvrecord
                    kvrecord_updated = False

                    try:
                        kvrecord = inspect_collection.data.query(
                            query=json.dumps({"_key": item_key})
                        )[0]

                        # update the kvrecord object_state, status_message and anomaly_reason
                        kvrecord["object_state"] = item_object_state
                        kvrecord["status_message"] = item.get("status_message")
                        kvrecord["anomaly_reason"] = item_anomaly_reason
                        kvrecord["mtime"] = time.time()
                        kvrecord["latest_flip_time"] = time.time()
                        kvrecord["latest_flip_state"] = item_object_state

                        # process the KVstore record update
                        inspect_collection.data.update(item_key, json.dumps(kvrecord))

                        kvrecord_updated = True

                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, collection record object update successfully, object="{item_object}", object_id="{item_key}"'
                        )

                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, failed to update the KVstore record, object="{item_object}", collection_name="{collection_name}", exception="{str(e)}"'
                        )

                    # proceeed with next steps
                    if kvrecord_updated:
                        try:

                            # calculate disruption time if current_state is green and previous_state was red
                            if (
                                item_object_state == "green"
                                and collection_object_state == "red"
                            ):
                                try:
                                    disruption_time = round(
                                        (time.time() - collection_latest_flip_time),
                                        2,
                                    )
                                except Exception as e:
                                    disruption_time = 0

                            flip_timestamp = time.strftime(
                                "%d/%m/%Y %H:%M:%S",
                                time.localtime(time.time()),
                            )
                            flip_result = f'{flip_timestamp}, object="{item_object}" has flipped from previous_state="{collection_object_state}" to state="{item_object_state}" with anomaly_reason="{item_anomaly_reason}", previous_anomaly_reason="{collection_anomaly_reason}", disruption_time="{disruption_time}"'

                            flip_record = {
                                "timeStr": flip_timestamp,
                                "tenant_id": self.tenant_id,
                                "object": item_object,
                                "object_category": item_object_category,
                                "object_state": item_object_state,
                                "object_previous_state": collection_object_state,
                                "latest_flip_time": time.time(),
                                "latest_flip_state": item_object_state,
                                "anomaly_reason": item_anomaly_reason,
                                "result": flip_result,
                            }

                            # add event_id
                            flip_record["event_id"] = hashlib.md5(
                                json.dumps(flip_record).encode()
                            ).hexdigest()

                            trackme_gen_state(
                                index=tenant_indexes["trackme_summary_idx"],
                                sourcetype="trackme:flip",
                                source="flip_state_change_tracking",
                                event=json.dumps(flip_record),
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, TrackMe flipping event created successfully, record="{json.dumps(flip_record, indent=1)}"'
                            )

                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, object="{item_object}", task="{task_name}", task_instance_id={task_instance_id}, failed to generate a flipping state event with exception="{e}"'
                            )

                        #
                        # SLA metrics
                        #

                        # create a list for SLA metrics generation
                        sla_metrics_records = []

                        if item_object_state == "green":
                            object_num_state = 1
                        elif item_object_state == "red":
                            object_num_state = 2
                        elif item_object_state == "orange":
                            object_num_state = 3
                        elif item_object_state == "blue":
                            object_num_state = 4
                        else:
                            object_num_state = 5

                        # add to our list
                        sla_metrics_records.append(
                            {
                                "tenant_id": self.tenant_id,
                                "object_id": item_key,
                                "object": item_object,
                                "alias": item_alias,
                                "object_category": item_object_category,
                                "monitored_state": item_monitored_state,
                                "priority": item_priority,
                                "metrics_event": {"object_state": object_num_state},
                            }
                        )

                        # call the SLA gen metrics function
                        sla_metrics_gen_start = time.time()
                        try:
                            sla_metrics = trackme_sla_gen_metrics(
                                self.tenant_id,
                                tenant_indexes.get("trackme_metric_idx"),
                                sla_metrics_records,
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, function trackme_sla_gen_metrics success {sla_metrics}, run_time={round(time.time()-sla_metrics_gen_start, 3)}, no_entities={len(sla_metrics_records)}'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, function trackme_sla_gen_metrics failed with exception {str(e)}'
                            )

                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, no_delta_records="{len(delta_records_keys)}", run_time="{round((time.time() - inspectcollection_compare_records_start_time), 3)}", collection="{inspect_collection_name}"'
                )

                #
                # END comparison
                #

                # end subtask
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 3)}", task has terminated.'
                )

                #
                #
                #

                #
                # Call the trackme_register_tenant_component_summary
                #

                # Use threading to do an async call to the register summary without waiting for it to complete
                thread = threading.Thread(
                    target=self.register_component_summary_async,
                    args=(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        component,
                    ),
                )
                thread.start()
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, trackme_register_tenant_component_summary was requested.'
                )

        #
        # task: untracked_entities
        #

        #
        # splk-dsm - untracked entities
        #

        # context: this activity tracks and maintain state for untracked entities
        # untracked entities are entities which are entirely out of the scope of any trackers, and therefore not maintained otherwise

        task_instance_id = self.get_uuid()
        task_name = "untracked_entities"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        if vtenant_record.get("tenant_dsm_enabled") == True:
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dsm", task="{task_name}", task_instance_id={task_instance_id}, inspecting untracked entities now.'
            )

            # kwargs
            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            untracked_entities_count = 0

            untracked_entities_search = f"""\
                | inputlookup trackme_dsm_tenant_{self.tenant_id} | eval key=_key

                ``` target any entity that has not been updated since more than 15m ```
                | eval time_sec_since_inspection=now()-tracker_runtime
                | where ( time_sec_since_inspection>900 OR isnull(tracker_runtime) )

                ``` called the offline abstract macro version ```
                `trackme_dsm_tracker_abstract({self.tenant_id})`

                ``` collects latest collection state into the summary index ```
                | `trackme_collect_state("current_state_tracking:splk-dsm:{self.tenant_id}", "object", "{self.tenant_id}")`

                ``` output flipping change status if changes ```
                | trackmesplkgetflipping tenant_id="{self.tenant_id}" object_category="splk-dsm"

                ``` update the KVstore collection ```
                | `trackme_outputlookup(trackme_dsm_tenant_{self.tenant_id}, key)`

                ``` update the delay metric only ```
                | `trackme_mcollect(object, splk-dsm, "metric_name:trackme.splk.feeds.lag_event_sec=data_last_lag_seen", "tenant_id, object_category, object", "{self.tenant_id}")`

                ``` summarize job ```
                | stats count as report_entities_count, values(object) as objects by tenant_id
            """

            # run the main report, every result is a Splunk search to be executed on its own thread
            try:
                reader = run_splunk_search(
                    self.service,
                    untracked_entities_search,
                    kwargs_oneshot,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        untracked_entities_count += 1
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dsm", task="{task_name}", task_instance_id={task_instance_id}, entities_count="{len(item)}"'
                        )

                if untracked_entities_count == 0:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dsm", task="{task_name}", task_instance_id={task_instance_id}, there are no untracked entities currently.'
                    )

            except Exception as e:
                # Call the component register
                trackme_register_tenant_object_summary(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    "all",
                    report_name,
                    "failure",
                    time.time(),
                    str(time.time() - start),
                    str(e),
                    "-5m",
                    "now",
                )
                msg = f'task="{task_name}", task_instance_id={task_instance_id}, tenant_id="{self.tenant_id}", main search failed with exception="{str(e)}"'
                logging.error(msg)
                raise Exception(msg)

        #
        # splk-dhm - untracked entities
        #
        # context: this activity tracks and maintain state for untracked entities
        # untracked entities are entities which are entirely out of the scope of any trackers, and therefore not maintained otherwise

        if vtenant_record.get("tenant_dhm_enabled") == True:
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dhm", task="{task_name}", task_instance_id={task_instance_id}, inspecting untracked entities now.'
            )

            # kwargs
            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            untracked_entities_count = 0

            untracked_entities_search = f"""\
                | inputlookup trackme_dhm_tenant_{self.tenant_id} | eval key=_key

                ``` target any entity that has not been updated since more than 15m ```
                | eval time_sec_since_inspection=now()-tracker_runtime
                | where ( time_sec_since_inspection>900 OR isnull(tracker_runtime) )

                ``` called the offline abstract macro version ```
                `trackme_dhm_tracker_abstract({self.tenant_id})`

                ``` collects latest collection state into the summary index ```
                | `trackme_collect_state("current_state_tracking:splk-dhm:{self.tenant_id}", "object", "{self.tenant_id}")`

                ``` output flipping change status if changes ```
                | trackmesplkgetflipping tenant_id="{self.tenant_id}" object_category="splk-dhm"

                ``` update the KVstore collection ```
                | `trackme_outputlookup(trackme_dhm_tenant_{self.tenant_id}, key)`

                ``` update the delay metric only ```
                | `trackme_mcollect(object, splk-dhm, "metric_name:trackme.splk.feeds.lag_event_sec=data_last_lag_seen", "tenant_id, object_category, object", "{self.tenant_id}")`

                ``` summarize job ```
                | stats count as report_entities_count, values(object) as objects by tenant_id
            """

            # run the main report, every result is a Splunk search to be executed on its own thread

            try:
                reader = run_splunk_search(
                    self.service,
                    untracked_entities_search,
                    kwargs_oneshot,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        untracked_entities_count += 1
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dhm", task="{task_name}", task_instance_id={task_instance_id}, entities_count="{len(item)}"'
                        )

                if untracked_entities_count == 0:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-dhm", task="{task_name}", task_instance_id={task_instance_id}, there are no untracked entities currently.'
                    )

            except Exception as e:
                # Call the component register
                trackme_register_tenant_object_summary(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    "all",
                    report_name,
                    "failure",
                    time.time(),
                    str(time.time() - start),
                    str(e),
                    "-5m",
                    "now",
                )
                msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, main search failed with exception="{str(e)}"'
                logging.error(msg)
                raise Exception(msg)

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: duplicated_entities
        #

        task_instance_id = self.get_uuid()
        task_name = "duplicated_entities"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        # all components except splk-wlk and splk-cim
        #
        # context: this situation is not expected, but if we have duplicated entities, we need to verify and purge them

        # splk-wlk - duplicated entities
        #
        # context: this activity tracks for duplicated entities in the Workload component
        # under some rare circumstances, the Splunk scheduler logs may lack the user context, althrough we implement several safeties
        # if this happens, we need to verify and purge any duplicated entity with the system user context instead of the proper user context

        for component in ("dsm", "dhm", "mhm", "wlk", "flx"):

            if (
                vtenant_record.get(f"tenant_{component}_enabled") == True
                and vtenant_record.get("tenant_replica") == False
            ):
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, inspecting entities now.'
                )

                # kwargs
                kwargs_oneshot = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "output_mode": "json",
                    "count": 0,
                }

                duplicated_entities_count = 0
                duplicated_entities_list = []

                # specific search for wlk
                if component == "wlk":
                    duplicated_entities_search = remove_leading_spaces(
                        f"""\
                        | inputlookup trackme_wlk_tenant_{self.tenant_id} | eval keyid=_key
                        | fields keyid, account, app, user, savedsearch_name, object, last_seen
                        | eventstats count as dcount by account, app, savedsearch_name
                        | where dcount>1
                        | sort - 0 savedsearch_name, last_seen
                    """
                    )

                else:  # other components
                    duplicated_entities_search = remove_leading_spaces(
                        f"""\
                        | inputlookup trackme_{component}_tenant_{self.tenant_id} | eval keyid=_key
                        | sort 0 object
                        | eventstats count as dcount by object
                        | streamstats count as rank by object
                        | where dcount>1 
                        ``` handle rank if the duplicated is due to FIPS migration ```
                        | eval rank=if(len(keyid) == 64, 2, 1)
                        | where rank=1
                    """
                    )

                # run the main report, every result is a Splunk search to be executed on its own thread
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, executing search="{duplicated_entities_search}"'
                )
                try:
                    reader = run_splunk_search(
                        self.service,
                        duplicated_entities_search,
                        kwargs_oneshot,
                        24,
                        5,
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            duplicated_entities_count += 1
                            duplicated_entities_list.append(item.get("keyid"))
                            logging.warning(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, detected duplicated entity, keyid="{item.get("keyid")}", object="{item.get("object")}"'
                            )

                    if duplicated_entities_count == 0:
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, there are no duplicated entities currently.'
                        )

                except Exception as e:
                    # Call the component register
                    trackme_register_tenant_object_summary(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        "all",
                        report_name,
                        "failure",
                        time.time(),
                        str(time.time() - start),
                        str(e),
                        "-5m",
                        "now",
                    )
                    msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, main search failed with exception="{str(e)}"'
                    logging.error(msg)
                    raise Exception(msg)

                # process if needed
                if duplicated_entities_count > 0:

                    # target
                    if component == "dsm":
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dsm/write/ds_delete"

                    elif component == "dhm":
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dhm/write/dh_delete"

                    if component == "mhm":
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_mhm/write/mh_delete"

                    if component == "flx":
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_flx/write/flx_delete"

                    if component == "wlk":
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_wlk/write/wlk_delete"

                    # data
                    # turn duplicated_entities_list into a comma separated string

                    # update comment
                    if component == "wlk":
                        update_comment = "One or more duplicated entities were detected by the health tracker, this condition can happen when Splunk scheduler logs lack the user context, automated purge of these entities."
                    else:
                        update_comment = "One or more duplicated entities were detected by the health tracker, this condition is not expected and TrackMe needs to purge duplicates to avoid further issues."

                    duplicated_entities_list = ",".join(duplicated_entities_list)
                    post_data = {
                        "tenant_id": self.tenant_id,
                        "keys_list": duplicated_entities_list,
                        "deletion_type": "temporary",
                        "update_comment": update_comment,
                    }

                    try:
                        response = requests.post(
                            target_url,
                            headers={
                                "Authorization": header,
                                "Content-Type": "application/json",
                            },
                            data=json.dumps(post_data),
                            verify=False,
                            timeout=300,
                        )
                        msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, duplicated entities purge successful, results="{json.dumps(response.json(), indent=2)}"'
                        logging.info(msg)

                    except Exception as e:
                        msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, duplicated entities purge failed with exception="{str(e)}"'
                        logging.info(msg)

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: synchronize_trackers_attr
        #

        task_instance_id = self.get_uuid()
        task_name = "synchronize_trackers_attr"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        #
        # Verify and synchronize trackers attribute knowledge
        #

        def process_sync_tracker_component(component):
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, verify and synchronize trackers knowledge attributes.'
            )

            # get macros
            macros = self.service.confs["macros"]

            # Data collection
            tracker_collection_name = (
                f"kv_trackme_{component}_hybrid_trackers_tenant_{self.tenant_id}"
            )
            tracker_collection = self.service.kvstore[tracker_collection_name]

            # Get the tenant KVrecord
            tracker_records = tracker_collection.data.query()

            for tracker_record in tracker_records:
                record_knowledge_objects = json.loads(
                    tracker_record.get("knowledge_objects")
                )
                record_key = tracker_record.get("_key")

                logging.debug(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, record_knowledge_objects={record_knowledge_objects}'
                )

                tracker_root_constraint_macro = record_knowledge_objects["properties"][
                    0
                ].get("root_constraint_macro")
                tracker_root_constraint = record_knowledge_objects["properties"][0].get(
                    "root_constraint"
                )

                try:
                    macro_definition = macros[tracker_root_constraint_macro].content[
                        "definition"
                    ]
                except Exception as e:
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, failed to retrieve the definition for the macro={tracker_root_constraint_macro}, exception={str(e)}'
                    )
                    macro_definition = None

                if macro_definition:
                    if macro_definition != tracker_root_constraint:
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, updating macro with new definition={macro_definition}'
                        )
                        # update the dict
                        record_knowledge_objects["properties"][0][
                            "root_constraint"
                        ] = macro_definition
                        # update the record
                        tracker_record["knowledge_objects"] = json.dumps(
                            record_knowledge_objects, indent=2
                        )

                        try:
                            tracker_collection.data.update(
                                str(record_key), json.dumps(tracker_record)
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, tracker record was successfully updated.'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, tracker record failed to be updated, exception={str(e)}'
                            )

        # Main logic
        components = ["dsm", "dhm", "mhm"]
        for component in components:
            if vtenant_record.get("tenant_" + component + "_enabled") == True:
                process_sync_tracker_component(component)

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: check_trackers
        #

        task_instance_id = self.get_uuid()
        task_name = "check_trackers_definition"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        def check_trackers_definition(vtenant_record, component):
            logging.info(f"Checking tracker definitions for component: {component}")

            # Load the tracker collection associated with the component (source of truth)
            tracker_collection_name = (
                f"kv_trackme_{component}_hybrid_trackers_tenant_{self.tenant_id}"
            )
            tracker_collection = self.service.kvstore[tracker_collection_name]

            # Get all the tracker records
            tracker_records = tracker_collection.data.query()

            # Initialize empty sets for the reports and macros that should be in the vtenant_record
            truth_reports = set()
            truth_macros = set()

            for tracker_record in tracker_records:
                record_knowledge_objects = json.loads(
                    tracker_record.get("knowledge_objects", "{}")
                )

                # Collect the reports and macros from the tracker record's knowledge_objects
                truth_reports.update(record_knowledge_objects.get("reports", []))
                truth_macros.update(record_knowledge_objects.get("macros", []))

            # Load the current tenant hybrid objects from vtenant_record (destination)
            hybrid_objects_json = vtenant_record.get(
                f"tenant_{component}_hybrid_objects"
            )

            if hybrid_objects_json:
                # Load the JSON object from the hybrid_objects field
                hybrid_objects = json.loads(hybrid_objects_json)
            else:
                # If no existing hybrid_objects, initialize an empty structure
                hybrid_objects = {"reports": [], "macros": []}

            vtenant_reports = set(hybrid_objects.get("reports", []))
            vtenant_macros = set(hybrid_objects.get("macros", []))

            # Compare and find missing reports/macros in the vtenant_record
            missing_reports = truth_reports - vtenant_reports
            missing_macros = truth_macros - vtenant_macros

            # If there are any missing reports or macros, add them to the vtenant_record
            if missing_reports or missing_macros:
                logging.info(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, Missing reports: {missing_reports} or macros: {missing_macros} in vtenant_record.'
                )

                # Update the vtenant_record with missing reports and macros
                hybrid_objects["reports"] = list(vtenant_reports.union(truth_reports))
                hybrid_objects["macros"] = list(vtenant_macros.union(truth_macros))

                # Save the updated hybrid objects back to the vtenant_record
                vtenant_record[f"tenant_{component}_hybrid_objects"] = json.dumps(
                    hybrid_objects, indent=2
                )

                try:
                    self.service.kvstore["kv_trackme_virtual_tenants"].data.update(
                        str(vtenant_record["_key"]), json.dumps(vtenant_record)
                    )
                    logging.info(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, vtenant_record updated successfully.'
                    )
                except Exception as e:
                    logging.error(
                        f'tenant_id="{self.tenant_id}", instance_id={instance_id}, component="splk-{component}", task="{task_name}", task_instance_id={task_instance_id}, Failed to update vtenant_record, exception: {str(e)}'
                    )

        # Main logic
        components = ["dsm", "dhm", "mhm", "flx", "wlk"]
        for component in components:
            if vtenant_record.get(f"tenant_{component}_enabled"):
                check_trackers_definition(vtenant_record, component)

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: logical_groups
        #

        task_instance_id = self.get_uuid()
        task_name = "check_logical_groups"
        task_start = time.time()

        #
        # Verify Logical Groups:
        # - load the logical groups KVstore collection
        # - verify that for each member of the groups, the member can be found in in any of the dsm/dhm/mhm/flx KVstore collection as an actively monitoreed entity
        # - if not, purge the member from the group
        #

        def query_kvstore_for_object(member, collection_suffix):
            target_collection_name = (
                f"kv_trackme_{collection_suffix}_tenant_{self.tenant_id}"
            )
            target_collection = self.service.kvstore[target_collection_name]
            query_string = {
                "$and": [
                    {
                        "object": member,
                        "monitored_state": "enabled",
                    }
                ]
            }
            try:
                kvrecord = target_collection.data.query(query=json.dumps(query_string))[
                    0
                ]
                kvrecord_key = kvrecord.get("_key", None)
            except:
                kvrecord_key = None

            if kvrecord_key:
                logging.debug(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, member="{member}", found in KVstore collection="{target_collection_name}"'
                )
                return True

            return False

        if (
            vtenant_record.get("tenant_dsm_enabled") == True
            or vtenant_record.get("tenant_dhm_enabled") == True
            or vtenant_record.get("tenant_mhm_enabled") == True
            or vtenant_record.get("tenant_flx_enabled") == True
        ):
            # log start
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting to verify logical groups, any orphan logical group member will be purged automatically.'
            )

            # time counter
            logical_group_check_start = time.time()

            #
            # Logical groups collection records
            #

            logical_group_coll = self.service.kvstore[
                f"kv_trackme_common_logical_group_tenant_{self.tenant_id}"
            ]

            (
                logical_groups_coll_records,
                logical_groups_by_group_key_dict,
                logical_groups_by_group_name_list,
                logical_groups_by_member_dict,
                logical_groups_by_member_list,
            ) = get_logical_groups_collection_records(logical_group_coll)

            # log all returned from the function
            logging.debug(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, logical_groups_coll_records={json.dumps(logical_groups_coll_records, indent=2)}, logical_groups_by_group_key_dict={json.dumps(logical_groups_by_group_key_dict, indent=2)}, logical_groups_by_group_name_list={json.dumps(logical_groups_by_group_name_list, indent=2)}, logical_groups_by_member_dict={json.dumps(logical_groups_by_member_dict, indent=2)}, logical_groups_by_member_list={json.dumps(logical_groups_by_member_list, indent=2)}'
            )

            # loops through logical_groups_by_member_list if not empty, then check in each KVstore collection if we have a match
            logical_member_found = False
            logical_members_orphans = []

            # ensure logical_groups_by_member_list is a list
            if isinstance(logical_groups_by_member_list, str):
                logical_groups_by_member_list = [logical_groups_by_member_list]

            if len(logical_groups_by_member_list) > 0:

                #
                # Orphans
                #

                for member in logical_groups_by_member_list:

                    for tenant_setting, collection_suffix in [
                        ("tenant_dsm_enabled", "dsm"),
                        ("tenant_dhm_enabled", "dhm"),
                        ("tenant_mhm_enabled", "mhm"),
                        ("tenant_flx_enabled", "flx"),
                    ]:
                        if vtenant_record.get(tenant_setting) == True:
                            logical_member_found = query_kvstore_for_object(
                                member, collection_suffix
                            )
                            if logical_member_found:
                                break
                    if not logical_member_found:
                        logical_members_orphans.append(member)

                # log orphans
                logging.debug(
                    f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, logical_members_orphans={json.dumps(logical_members_orphans, indent=2)}'
                )

                # purge orphans
                if len(logical_members_orphans) > 0:
                    # turn the list into a comma separated string
                    logical_members_orphans = ",".join(logical_members_orphans)

                    try:
                        logical_group_purge_remove_response = (
                            logical_group_remove_object_from_groups(
                                self._metadata.searchinfo.splunkd_uri,
                                self._metadata.searchinfo.session_key,
                                self.tenant_id,
                                logical_members_orphans,
                            )
                        )
                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, orphan_members="{logical_members_orphans}", successfully purged the logical groups collection, response="{json.dumps(logical_group_purge_remove_response, indent=2)}"'
                        )

                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, orphan_members="{logical_members_orphans}", failed to purge from the logical groups collection, exception={str(e)}'
                        )

                #
                # empty groups
                #

                for logical_group_record in logical_groups_coll_records:

                    # get the group name
                    object_group_name = logical_group_record.get("object_group_name")

                    # get the members
                    members = logical_group_record.get("object_group_members", None)
                    if members:
                        if not len(members) > 0:
                            members = None

                    if not members:

                        logging.info(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, group="{object_group_name}", group has no members, will be purged.'
                        )

                        try:
                            logical_group_delete_response = (
                                logical_group_delete_group_by_name(
                                    self._metadata.searchinfo.splunkd_uri,
                                    self._metadata.searchinfo.session_key,
                                    self.tenant_id,
                                    object_group_name,
                                )
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, group="{object_group_name}", group has been purged successfully, response="{json.dumps(logical_group_delete_response, indent=2)}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, group="{object_group_name}", failed to purge the group, exception={str(e)}'
                            )

            # log time
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, logical_groups_check_duration="{round(time.time() - logical_group_check_start, 3)}"'
            )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: check_trackers
        #

        task_instance_id = self.get_uuid()
        task_name = "check_trackers_statuses"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        # Set the query
        health_search = remove_leading_spaces(
            f"""
        | trackme mode=post url=/services/trackme/v2/configuration/get_tenant_ops_status body=\"{{'mode': 'raw', 'tenant_id': '{self.tenant_id}'}}\"
        | trackmeopsstatusexpand
        """
        )

        # logging
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, tenant_id="{self.tenant_id}", Starting health_search, report="{report_name}", search="{health_search}"'
        )

        # kwargs
        kwargs_oneshot = {
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }

        # run the main report, every result is a Splunk search to be executed on its own thread
        try:
            reader = run_splunk_search(
                self.service,
                health_search,
                kwargs_oneshot,
                24,
                5,
            )

            # Call the component register
            trackme_register_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                "all",
                report_name,
                "success",
                time.time(),
                str(time.time() - start),
                "The report was executed successfully",
                "-5m",
                "now",
            )

            for item in reader:
                if isinstance(item, dict):
                    # verify the knowledge object - if for some reason it is not existing anymore, we should remove it
                    # and not take it into account any longer

                    # process
                    savedsearch_definition = None
                    report_name = item.get("report")
                    try:
                        savedsearch = self.service.saved_searches[report_name]
                        savedsearch_definition = savedsearch.content["search"]
                    except Exception as e:
                        savedsearch_definition = None
                        logging.warning(
                            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, the report="{report_name}" does not exist anymore, somehow it was removed without TrackMe being aware of it, will get rid of this now.'
                        )

                    if not savedsearch_definition:
                        # extract component
                        component = report_name.split("_")[1]

                        # purge
                        try:
                            delete_register_summary = (
                                trackme_delete_tenant_object_summary(
                                    self._metadata.searchinfo.session_key,
                                    self._metadata.searchinfo.splunkd_uri,
                                    self.tenant_id,
                                    f"splk-{component}",
                                    report_name,
                                )
                            )
                            logging.info(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, knowledge for the report="{report_name}" was purged successfully, response="{delete_register_summary}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, exception encountered while calling function trackme_delete_tenant_object_summary, exception="{str(e)}"'
                            )

                    else:
                        yield {
                            "_time": time.time(),
                            "_raw": item,
                            "component": item.get("component"),
                            "earliest": item.get("earliest"),
                            "last_duration": item.get("last_duration"),
                            "last_exec": item.get("last_exec"),
                            "last_result": item.get("last_result"),
                            "last_status": item.get("last_status"),
                            "latest": item.get("latest"),
                            "report": report_name,
                            "tenant_id": item.get("tenant_id"),
                        }

                        # index the audit record
                        try:
                            trackme_state_event(
                                session_key=self._metadata.searchinfo.session_key,
                                splunkd_uri=self._metadata.searchinfo.splunkd_uri,
                                tenant_id=self.tenant_id,
                                index=tenant_indexes["trackme_audit_idx"],
                                sourcetype="trackme:health",
                                source=f"trackme:health:{self.tenant_id}",
                                record={
                                    "component": item.get("component"),
                                    "earliest": item.get("earliest"),
                                    "last_duration": item.get("last_duration"),
                                    "last_exec": item.get("last_exec"),
                                    "last_result": item.get("last_result"),
                                    "last_status": item.get("last_status"),
                                    "latest": item.get("latest"),
                                    "report": report_name,
                                    "tenant_id": item.get("tenant_id"),
                                },
                            )
                        except Exception as e:
                            error_msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, exception encountered while calling function trackme_state_event, exception="{str(e)}"'
                            logging.error(error_msg)
                            raise Exception(error_msg)

        except Exception as e:
            # Call the component register
            trackme_register_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                "all",
                report_name,
                "failure",
                time.time(),
                str(time.time() - start),
                str(e),
                "-5m",
                "now",
            )
            msg = f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, main search failed with exception="{str(e)}"'
            logging.error(msg)
            raise Exception(msg)

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        #
        # task: apply_licensing_restrictions
        #

        task_instance_id = self.get_uuid()
        task_name = "apply_licensing_restrictions"
        task_start = time.time()

        # start task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, starting task.'
        )

        #
        # licensing restriction
        #

        # if the component is a restricted component and the product is not registered, it should be disabled now
        if license_is_valid != 1 and (
            vtenant_record.get("tenant_cim_enabled") == 1
            or vtenant_record.get("tenant_flx_enabled") == 1
            or vtenant_record.get("tenant_wlk_enabled") == 1
        ):
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, due to licensing restrictions, this tenant will be automatically disabled, the tenant is running a restricted component while this instance is not registered'
            )

            # target
            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

            # data
            post_data = {
                "tenant_id": self.tenant_id,
                "update_comment": "Auto disabling this tenant due to licensing limitation, the tenant is running a restricted component while the product is not currently registered",
                "force": "true",
            }

            try:
                response = requests.post(
                    target_url,
                    headers={
                        "Authorization": header,
                        "Content-Type": "application/json",
                    },
                    data=json.dumps(post_data),
                    verify=False,
                    timeout=300,
                )
                return json.loads(response.text)

            except Exception as e:
                raise Exception(
                    f'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{str(e)}"'
                )

        elif (
            license_is_valid != 1
            and license_active_tenants > 2
            and self.tenant_id not in license_active_tenants_list[0:2]
        ):
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, due to licensing restrictions, this tenant will be automatically disabled, the tenant is running a restricted component while this instance is not registered'
            )

            # target
            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

            # data
            post_data = {
                "tenant_id": self.tenant_id,
                "update_comment": f"Auto disabling this tenant due to licensing limitation, this deployment has reached the maximum number of tenants allowed ({license_active_tenants}), only the following tenants can be used: {license_active_tenants_list[0:2]}",
                "force": "true",
            }

            try:
                response = requests.post(
                    target_url,
                    headers={
                        "Authorization": header,
                        "Content-Type": "application/json",
                    },
                    data=json.dumps(post_data),
                    verify=False,
                    timeout=300,
                )
                return json.loads(response.text)

            except Exception as e:
                raise Exception(
                    f'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{str(e)}"'
                )

        elif (
            license_is_valid == 1
            and license_subscription_class == "enterprise"
            and license_active_tenants > 6
            and self.tenant_id not in license_active_tenants_list[0:6]
        ):
            logging.info(
                f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, due to licensing restrictions, this tenant will be automatically disabled, the tenant is over the maximum number of allowed tenants in Enterprise Edition'
            )

            # target
            target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

            # data
            post_data = {
                "tenant_id": self.tenant_id,
                "update_comment": f"Auto disabling this tenant due to licensing limitation, this deployment has reached the maximum number of tenants allowed ({license_active_tenants}), only the following tenants can be used: {license_active_tenants_list[0:6]}",
                "force": "true",
            }

            try:
                response = requests.post(
                    target_url,
                    headers={
                        "Authorization": header,
                        "Content-Type": "application/json",
                    },
                    data=json.dumps(post_data),
                    verify=False,
                    timeout=300,
                )
                return json.loads(response.text)

            except Exception as e:
                raise Exception(
                    f'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{str(e)}"'
                )

        # end task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, task="{task_name}", task_instance_id={task_instance_id}, run_time="{round(time.time()-task_start, 2)}", task has terminated.'
        )

        # end general task
        logging.info(
            f'tenant_id="{self.tenant_id}", instance_id={instance_id}, trackmetrackerhealth has terminated, total_run_time={round(time.time() - start, 3)}'
        )


dispatch(HealthTracker, sys.argv, sys.stdin, sys.stdout, __name__)
